# CubiCal: a radio interferometric calibration suite
# (c) 2017 Rhodes University & Jonathan S. Kenyon
# http://github.com/ratt-ru/CubiCal
# This code is distributed under the terms of GPLv2, see LICENSE.md for details
from __future__ import print_function
from cubical.machines.interval_gain_machine import PerIntervalGains
import numpy as np
from scipy import special, stats
from cubical.flagging import FL
import cubical.kernels
import time

from cubical.tools import logger, ModColor
log = logger.getLogger("robust_2x2")  #TODO check this "complex_2x2"

class ComplexW2x2Gains(PerIntervalGains):
    """
    This class implements the weighted full complex 2x2 gain machine based on the Complex T-distribution
    """
    
    def __init__(self, label, data_arr, ndir, nmod,
                 chunk_ts, chunk_fs, chunk_label, options):

        """
        Initialises a weighted complex 2x2 gain machine.
        
        Args:
            label (str):
                Label identifying the Jones term.
            data_arr (np.ndarray): 
                Shape (n_mod, n_tim, n_fre, n_ant, n_ant, n_cor, n_cor) array containing observed 
                visibilities. 
            ndir (int):
                Number of directions.
            nmod (nmod):
                Number of models.
            chunk_ts (np.ndarray):
                Times for the data being processed.
            chunk_fs (np.ndarray):
                Frequencies for the data being processsed.
            options (dict): 
                Dictionary of options. 
        """

        PerIntervalGains.__init__(self, label, data_arr, ndir, nmod, chunk_ts, chunk_fs, chunk_label, options)

        if self.is_diagonal:
            self.kernel_robust = cubical.kernels.import_kernel("diag_robust")
        else:
            self.kernel_robust = cubical.kernels.import_kernel("full_W_complex")

        self.residuals = np.empty_like(data_arr)

        self.save_weights = options.get("robust-save-weights", False)
        
        self.label = label

        self.cov_type = options.get("robust-cov", "compute") # adding an option to compute residuals covariance or just assume 1 as in Robust-t paper

        self.npol = options.get("robust-npol", 2.) # testing if the number of polarizations really have huge effects

        self.v_int = options.get("robust-int", 5)

        self.cov_scale = options.get("robust-scale", 0) # scale down the covariance by this factor

        self.cov_thresh = options.get("robust-cov-thresh",  1)

        self.sigma_thresh = options.get("robust-sigma-thresh", 3)

        self.robust_flag_weights = options.get("robust-flag-weights", True)

        self.fixed_v = False

        self.not_all_flagged = True

        self._flag = True

        self.any_new = 0

        self.is_robust = True # to identify this machine as the robust solver

        self._estimate_pzd = options["estimate-pzd"]

        # this will be set to PZD and exp(-i*PZD) once the PZD estimate is done
        self._pzd = 0
        self._exp_pzd = 1


    @classmethod
    def determine_diagonality(cls, options):
        """Returns true if the machine class, given the options, represents a diagonal gain"""
        updates = set(options['update-type'].split("-"))
        return options['type'] == 'robust-diag' or \
               ('full' not in updates and 'leakage' not in updates)

    @staticmethod
    def exportable_solutions():
        """ Returns a dictionary of exportable solutions for this machine type. """
        sols = PerIntervalGains.exportable_solutions()
        sols["pzd"] = (0.0, ("time", "freq"))
        return sols

    def importable_solutions(self, grid0):
        """ Returns a dictionary of importable solutions for this machine type. """
        sols = super(ComplexW2x2Gains, self).importable_solutions(grid0)
        if "pzd" in self.update_type:
            sols["pzd"] = dict(**self.interval_grid)
        return sols

    def export_solutions(self):
        """ Saves the solutions to a dict of {label: solutions,grids} items. """
        sols = super(ComplexW2x2Gains, self).export_solutions()
        if "pzd" in self.update_type and self._pzd is not None:
            sols["pzd"] = (masked_array(self._pzd), self.interval_grid)
        return sols

    def import_solutions(self, soldict):
        """
        Loads solutions from a dict.

        Args:
            soldict (dict):
                Contains gains solutions which must be loaded.
        """
        if "pzd" in self.update_type and "pzd" in soldict:
            self._pzd = soldict["pzd"]
            self._exp_pzd = np.exp(-1j * self._pzd)

        super(ComplexW2x2Gains, self).import_solutions(soldict)

    def compute_js(self, obser_arr, model_arr):
        """
        This function computes the (J^H)WR term of the weighted GN/LM method for the
        full-polarisation, phase-only case.

        Args:
            obser_arr (np.array): Array containing the observed visibilities.
            
            model_arr (np.array): Array containing the model visibilities.
            
            gains (np.array): Array containing the current gain estimates.

        Returns:
            Returns:
            
            jhwr (np.array): Array containing the result of computing (J^H)WR.
            
            jhwjinv (np.array): Array containing the result of computing (J^HW.J)^-1
            
            flag_count:     Number of flagged (ill-conditioned) elements
        """
        # w = self.weights
        
        n_dir, n_tim, n_fre, n_ant, n_cor, n_cor = self.gains.shape

        # compute residuals and weights
        if self.iters == 1:
            self.residuals = self.compute_residual(obser_arr, model_arr, self.residuals)

            if self.not_all_flagged:
                self.update_weights()


        jh = self.get_new_jh(model_arr)

        if self.offdiag_only:
            jh[...,(0,1),(0,1)] = 0
            self.residuals[...,(0,1),(0,1)] = 0
            
        if self.diag_only:
            jh[...,(0,1),(1,0)] = 0
            self.residuals[...,(0,1),(1,0)] = 0

        self.kernel_robust.compute_jh(model_arr, self.gains, jh, self.t_int, self.f_int)

        jhwr = self.get_new_jhr()

        self.kernel_robust.compute_jhwr(jh, self.residuals, self.weights, jhwr, self.t_int, self.f_int) #TODO 

        jhwj, jhwjinv = self.get_new_jhj()

        self.kernel_robust.compute_jhwj(jh, self.weights, jhwj, self.t_int, self.f_int)

        flag_count = self.kernel_robust.compute_jhjinv(jhwj, jhwjinv, self.gflags, self.eps, FL.ILLCOND)
        
        return jhwr, jhwjinv, flag_count

    
    #@profile
    def implement_update(self, jhr, jhjinv):

        # jhjinv is 2x2 block-diagonal, with Hermitian blocks. TODO: what's the variance on the off-diagonals?
        # variance of gain is diagonal of jhjinv
        # not sure about how the  weights affects the posterior variance. here we actually pass jhwj, not jhj
        if self.posterior_gain_error is None:
            self.posterior_gain_error = np.zeros_like(jhjinv.real)
        
        #----normalising to reduce the effects of the weights on jhj---#
        #----not sure if this is the best way to do this---------------#

        self.Nvis = np.sum(self.new_flags==False)
        try:
            norm = np.min(self.weights[np.where(self.weights>0)]).real #np.real((1/2.)*np.sum(self.weights)) #/self.Nvis
        except:
            norm = 0
        diag = norm*jhjinv[..., (0, 1), (0, 1)].real #/np.max(self.weights)
        self.posterior_gain_error[...,(0,1),(0,1)] = np.sqrt(diag)
        self.posterior_gain_error[...,(1,0),(0,1)] = np.sqrt(diag.sum(axis=-1)/2)[...,np.newaxis]

        update = self.init_update(jhr)
        self.kernel_robust.compute_update(jhr, jhjinv, update)


        # if self.dd_term and self.n_dir > 1: computing residuals for both DD and DID calibration
        update += self.gains
        self.restrict_solution(update)

        if self.iters % 2 == 0 or self.n_dir > 1:
            self.gains += update
            self.gains *= 0.5
            self.restrict_solution(self.gains)
        else:
            np.copyto(self.gains, update)


    def compute_update(self, model_arr, obser_arr):
        """
        This method is expected to compute the parameter update. 
        
        The standard implementation simply calls compute_js() and implement_update(), here we are
        overriding it because we need to update weights as well. 

        Args:
            model_arr (np.ndarray): 
                Shape (n_dir, n_mod, n_tim, n_fre, n_ant, n_ant, n_cor, n_cor) array containing 
                model visibilities. 
            
            obser_arr (np.ndarray): 
                Shape (n_mod, n_tim, n_fre, n_ant, n_ant, n_cor, n_cor) array containing observed 
                visibilities. 
        """

        jhwr, jhwjinv, flag_count = self.compute_js(obser_arr, model_arr)

        self.implement_update(jhwr, jhwjinv)

        # Computing the weights

        if self.not_all_flagged:
            self.residuals = self.compute_residual(obser_arr, model_arr, self.residuals)
            self.update_weights()

        return flag_count

    def compute_covinv(self):
        
        """
        This functions computes the 4x4 covariance matrix of the residuals visibilities, 
        and it approximtes it inverse. I self.cov_type is set to 1, the covariance maxtrix is 
        assumed to be the Identity matrix as in the Robust-t paper.

        Args:
            residuals (np.array) : Array containing the residuals.
                Shape is n_dir, n_tim, n_fre, n_ant, a_ant, n_cor, n_cor

        Returns:
            covinv (np.array) : Shape is ncor*n_cor x ncor*n_cor (4x4)
            Array containing the inverse covariance matrix

        """

        if self.cov_type == "identity":

            covinv = np.eye(4, dtype=self.dtype)
          
        else:

            Nvis = self.Nvis/2. #only half of the visibilties are used for covariance computation

            ompstd = np.zeros((4,4), dtype=self.dtype)

            self.kernel_robust.compute_cov(self.residuals, ompstd, self.weights)

            # removing the offdiagonal correlations
            std = np.diagonal(ompstd/Nvis) + self.eps**2 # To avoid division by zero

            if np.any(std>self.cov_thresh):
                self.fixed_v = True
                
                if self.flaground:
                    print(ModColor.Str("rb-2x2 {} : flag round {}: Warning Covariance too high probably because of RFI will fixed v to 2 and cov to 1".format(self.label, self._count+1), "red"), file=log(2))
                else:
                    print(ModColor.Str("rb-2x2 {} : {} iters: Warning Covariance too high probably because of RFI will fixed v to 2 and cov to 1".format(self.label, self.iters), "red"), file=log(2))

            else:
                #---scaling the variance in this case improves the robust solver performance----------#
                self.fixed_v = False
                if self.cov_scale and not self.flaground:
                    std /= self.cov_scale 

            if self.iters % 5 == 0 or self.iters == 1:
                if self.flaground:
                    print("rb-2x2 {} : flag round {}: covariance diagonal : [{}]".format(self.label, self._count+1, ", ".join('{:.2g}'.format(x) for x in std)), file=log(2))
                    # print("rb-2x2 {} : flag round {}: covariance diagonal : [{}]".format(self.label, self._count+1, ", ".join('{:.2g}'.format(x) for x in std2)), file=log(2))
                else:
                    print("rb-2x2 {} : {} iters: covariance diagonal : [{}]".format(self.label, self.iters, ", ".join('{:.2g}'.format(x) for x in std)), file=log(2))
                    # print("rb-2x2 {} : {} iters: covariance diagonal : [{}]".format(self.label, self.iters, ", ".join('{:.2g}'.format(x) for x in std2)), file=log(2))


            # can we disable the flagging the solver to avoid flagging unmodelled sources
            # UMS: my thought here is that if data is unmodelled sources rather than RFI xx and yy covariance should be very close
            # so the solver should disable the flagging in this case
            xx_close_to_yy = 0.8 <= np.abs(std[0])/np.abs(std[0]) <= 1.2
            cov_low = np.average([std[0], std[3]]).real < 2e-2

            if xx_close_to_yy and cov_low and self.flaground and self._count==0:
                self.robust_flag_disable = True
                print(ModColor.Str("rb-2x2 {} : flag round {}: Warning: the covariance is low and the xx and yy variances are very close. Flagging will be disable".format(self.label, self._count+1), "red"), file=log(2))


            covinv = np.eye(4, dtype=self.dtype)

            if self.fixed_v:
                covinv *= 1/self.cov_thresh
            else:
                if self.cov_type == "hybrid":
                    if np.max(np.abs(std)) < 1:
                        covinv[np.diag_indices(4)]= 1/std
                
                elif self.cov_type == "compute":
                    covinv[np.diag_indices(4)]= 1/std
                else:
                    raise RuntimeError("unknown robust-cov setting")
        
        if self.npol == 2:
            covinv[(1,2), (1,2)] = 0


        self.covinv = covinv

        # print("rb-2x2 {} : {} iters: hybrid check covariance diagonal : [{}]".format(self.label, self.iters, ", ".join('{:.2g}'.format(x) for x in np.diagonal(covinv))), file=log(2))


    def update_weights(self):
        
        """
            This computes the weights, given the latest residual visibilities and the v parameter.
            w[i] = (v+2*npol)/(v + 2*r[i].T.cov.r[i]. Next v is update using the newly compute weights.
        
            Args:
                r (np.array): Array of the residual visibilities.
                covinc (np.array) : Array containing the inverse of 
               covariance of residual visibilities
                w (np.array): Array containing the weights
                v (float) : v -- number of degrees of freedom

            Returns:
                w (np.array) : new weights
                v (float) : new value for v
        """

        def  _brute_solve_v(wn):
            """Finds a root for the function f constraint between low (2) and high (50)
            Args:
                wn : the weights flatenncc-a-b in to a 1D array
            Returns:
                root (float) : The root of f or minimum point    
            """
            # import pdb; pdb.set_trace()

            m = len(wn)

            
            vfunc = lambda a: special.digamma(a+self.npol) - np.log(a+self.npol) - \
                        special.digamma(a) + np.log(a) + (1./m)*np.sum(np.log(wn) - wn) + 1
            

            vvals = np.arange(2, 101, 1, dtype=float) #search for v in the range (2, 100)
            fvals = vfunc(vvals)
            root = vvals[np.argmin(np.abs(fvals))]

            if self.flaground:
                print("rb-2x2 {} : flag round {}: v-parameter is  {:.3}".format(self.label, self.iters, root), file=log(2))
                print("rb-2x2 {} : flag round {} : weights stats are min {:.4}, max: {:.4}, mean: {:.4}, std: {:.4}, median: {:.4}, mad: {:.4}".format(self.label, self._count+1, np.min(wn), 
                                                    np.max(wn), np.mean(wn), np.std(wn), np.median(wn), stats.median_absolute_deviation(wn)), file=log(2)) 
            else:
                
                if self.iters % 5 == 0 or self.iters == 1:
                    print("rb-2x2 {} : {} iters: v-parameter is  {:.3}".format(self.label, self.iters, root), file=log(2))
                    #nless, nmore = len(wn[wn<1])/len(wn), len(wn[wn>1])/len(wn)
                    print("rb-2x2 {} : {} iters: weights stats are min {:.4}, max: {:.4}, mean: {:.4}, std: {:.4}, median: {:.4}, mad: {:.4}".format(self.label, self.iters, np.min(wn), 
                                                    np.max(wn), np.mean(wn), np.std(wn), np.median(wn), stats.median_absolute_deviation(wn)), file=log(2)) #, np.sum(wn), m, nless, nmore, 
            
            if self.fixed_v:
                return 2
            else:    
                return root

        if self.iters % self.v_int == 0 or self.iters == 1:
            self.compute_covinv()

        if self.fixed_v:
            self.v = 2

        self.kernel_robust.compute_weights(self.residuals, self.covinv, self.weights, self.v, self.npol)

        # re-set weights for visibillities flagged from start to 0 and normalise the weights by their sum 
        self.weights[:,self.new_flags!=0,:] = 0
        
        aa, ab = np.tril_indices(self.n_ant, -1)
        w_real = np.real(self.weights[:,:,:,aa,ab,0].flatten())
        w_nzero = w_real[np.where(w_real!=0)[0]]
        norm = np.average(w_nzero)

        self.weights/= norm
        w_nzero/=norm
        
        # wstd = np.std(w_nzero) #stats.median_absolute_deviation(w_nzero)
        # wmean = 1 #np.median(w_nzero)
        
        #-----------computing the v parameter---------------------#
        # This computation is only done after a certain number of iterations. Default is 5
        if self.iters % self.v_int == 0 or self.iters == 1:
            self.v = _brute_solve_v(w_nzero) # remove zero weights

        self.not_all_flagged = self.flag_weights()
        
        return

    def flag_weights(self):
        """Trying flagging visiblities with very very low weights and see if this improves the solution
        like mad max flagger

        wstd: the weights standard deviation
        """

        if self.flaground and not self.robust_flag_disable:

            if self._final_flaground:
                _pre_or_post = "after-solving"   
            else:
                _pre_or_post = "before solving"

            _nvis = self.new_flags.size
            nflag0 = np.sum(self.new_flags!=0)
            wfrac0 = nflag0/_nvis

            # This correction factor is two ensure that we only flag when v is low.
            # The 1.26 factor is just makes it work somehow 
            _v_corr = 1.26/self.v #2*self.npol/self.v if self.v < 5 else self.npol/self.v

            wlow =  _v_corr*(self.v + self.npol)/(self.v + self.sigma_thresh)

            # print("rb-2x2 {} : {} iters: wlow is  {:.3} while 1/v is {:.3}".format(self.label, self.iters, wlow, 1/self.v), file=log(2))

            self.weight_flags = np.where((self.weights< wlow) & (self.weights!=0)) # wlow

            # import pdb; pdb.set_trace()

            if len(self.weight_flags[0])>0:

                self.weights[self.weight_flags] = 0
                self.new_flags[self.weight_flags[1:-1]] |= FL.MAD
                self.residuals[self.weight_flags[:-1]] = 0

                nflag = np.sum(self.new_flags!=0)

                any_new = nflag-nflag0

                if any_new:
                    wfrac = any_new/_nvis
                    
                    print(ModColor.Str("rb-2x2 {} : {} flag round {} : number of weight flags {} ({:.4%}), prior flags {} ({:.4%})".format(self.label, _pre_or_post, self._count+1, any_new, wfrac, nflag0, wfrac0), "blue"), file=log(2))
                
                self.any_new = True

                self._count += 1 
                
                return False if nflag ==_nvis else True

            else:
                if self._count==0:
                    self.any_new = False 
                return True
        
        else:
            self.any_new = False
            self.weight_flags = None 
            
            return True


    def update_weight_flags(self, flags_arr):

        self.new_flags = flags_arr
        self.weights[:,self.new_flags!=0,:] = 0
        
        self.Nvis = np.sum(self.new_flags==False)

        self.not_all_flagged = False if self.Nvis==0 else True

    def robust_flag(self, flags_arr, model_arr, obser_arr, final=False):
        """run an iteration and use the weights to flag high level RFIs"""

        self.flaground = True

        if final is False:
            self.v = 5  # Don't start with a low degree of freedom to prevent overflagging
            self.iters = 1

        self._final_flaground = True if final else False

        if final and self.not_all_flagged:
            self.flag_weights()
        else:
            self.compute_update(model_arr, obser_arr)
            np.copyto(self.gains, self.old_gains)

        if self.any_new:
            flags_arr[self.weight_flags[1:-1]] |= FL.MAD
            self.update_equation_counts(flags_arr != 0)

            new_flags = flags_arr & ~(FL.MISSING | FL.PRIOR) != 0
                
            model_arr[:, :, new_flags!=0, :, :] = 0
            obser_arr[   :, new_flags!=0, :, :] = 0

        if final is False:
            self.weights[:, self.new_flags==0, :] = 1
            self.iters = 0
            self.v = 2
            self._gh_update = self._ghinv_update = True

        self.flaground = False

    def restrict_solution(self, gains):
        
        
        if "pzd" in self.update_type:
            # re-estimate pzd
            mask = self.gflags!=0
            with np.errstate(divide='ignore', invalid='ignore'):
                pzd = masked_array(gains[:, :, :, :, 0, 0] / gains[:, :, :, :, 1, 1], mask)
            pzd = np.angle(pzd.sum(axis=(0,3)))
            with np.printoptions(precision=4, suppress=True, linewidth=1000):
                print("{0}: PZD estimate changes by {1} deg".format(self.chunk_label, (pzd-self._pzd)* 180 / np.pi), file=log(2))
            # import ipdb; ipdb.set_trace()
            self._pzd = pzd
            self._exp_pzd = np.exp(-1j * pzd)

            gains[:, :, :, :, 0, 0] = 1
            gains[:, :, :, :, 1, 1] = self._exp_pzd[np.newaxis, :, :, np.newaxis]
            
        if "leakage" in self.update_type:
            if "pzd" not in self.update_type:
                gains[:, :, :, :, 0, 0] = 1
                gains[:, :, :, :, 1, 1] = 1
                
            if "rel" in self.update_type and self.ref_ant is not None:
                offset =  gains[:, :, :, self.ref_ant, 0, 1].copy()
                gains[..., 0, 1] -= offset[..., np.newaxis]
                gains[..., 1, 0] += np.conj(offset)[..., np.newaxis]
                with np.printoptions(precision=4, suppress=True, linewidth=1000):
                    print("{0}: subtracting relative leakage offset {1}".format(self.chunk_label, offset), file=log(2))


        if self.ref_ant is not None:
            phase = np.angle(self.gains[...,self.ref_ant,0,0])
            gains[:,:,:,:,(0,1),(0,1)] *= np.exp(-1j*phase)[:,:,:,np.newaxis,np.newaxis]

        super(ComplexW2x2Gains, self).restrict_solution(gains)


    def precompute_attributes(self, data_arr, model_arr, flags_arr, noise):
        """
        Set the initial weights to 1 and set the weights of the flags data points to 0

        Args:
            model_arr (np.ndarray):
                Shape (n_dir, n_mod, n_tim, n_fre, n_ant, n_ant, n_cor, n_cor) array containing 
                model visibilities.
            flags_arr (np.ndarray):
                Shape (n_tim, n_fre, n_ant, n_ant) array containing  flags
        """
        
        super(ComplexW2x2Gains, self).precompute_attributes(data_arr, model_arr, flags_arr, noise)

        if self._estimate_pzd and self._pzd is 0:
            marr = model_arr[..., (0, 1), (1, 0)][:, 0].sum(0)
            darr = data_arr[..., (0, 1), (1, 0)][0]
            mask = (flags_arr[..., np.newaxis] != 0) | (marr == 0)
            with np.errstate(divide='ignore', invalid='ignore'):
                dm = darr * (np.conj(marr) / abs(marr))
            dabs = np.abs(darr)
            dm[mask] = 0
            dabs[mask] = 0
            # collapse time/freq axis into intervals and sum antenna axes
            dm_sum = self.interval_sum(dm).sum(axis=(2, 3))
            dabs_sum = self.interval_sum(dabs).sum(axis=(2, 3))
            # sum off-diagonal terms
            dm_sum = dm_sum[..., 0] + np.conj(dm_sum[..., 1])
            dabs_sum = dabs_sum[..., 0] + np.conj(dabs_sum[..., 1])
            pzd = np.angle(dm_sum / dabs_sum)
            pzd[dabs_sum == 0] = 0
            with np.printoptions(precision=4, suppress=True, linewidth=1000):
                print("{0}: PZD estimate {1} deg".format(self.chunk_label, pzd * 180 / np.pi), file=log(2))
            self._pzd = pzd
            self._exp_pzd = np.exp(-1j * pzd)

            self.gains[:, :, :, :, 0, 0] = 1
            self.gains[:, :, :, :, 1, 1] = self._exp_pzd[np.newaxis, :, :, np.newaxis]

        self.weights_shape = [self.n_mod, self.n_tim, self.n_fre, self.n_ant, self.n_ant, 1]
        
        self.weights = np.ones(self.weights_shape, dtype=self.dtype)

        self.v = 2 # t-distribution number of degrees of freedom

        self.covinv = np.eye(4, dtype=self.dtype)

        self.weight_flags = None

        self.update_weight_flags(flags_arr)

        self.flaground = False

        self._final_flaground = False

        self.robust_flag_disable = False

        self._count = 0

    @property
    def dof_per_antenna(self):
        if "leakage" in self.update_type and "pzd" in self.update_type:
            return 4 + 1./self.n_ant
        elif "leakage" in self.update_type:
            return 4
        elif "pzd" in self.update_type:
            return 1./self.n_ant
        else:
            return super(ComplexW2x2Gains, self).dof_per_antenna



# CubiCal: a radio interferometric calibration suite
# (c) 2017 Rhodes University & Jonathan S. Kenyon
# http://github.com/ratt-ru/CubiCal
# This code is distributed under the terms of GPLv2, see LICENSE.md for details

# This module has been adapted from the DDFacet package,
# (c) Cyril Tasse et al., see http://github.com/saopicc/DDFacet
from __future__ import print_function
from builtins import range
import os
import sys
from . import ModColor

class ClassPrint():
    def __init__(self,HW=20,quote='"'):
        self.LeftW=HW
        self.WV0=50
        self.WV1=30
        self.WidthHelp=30
        self.quote=quote
        self.proto=" - %s %s %s : "
        self.Lproto=len(self.proto.replace("%s",""))


    def getWidth(self):
        rows, columns = os.popen('stty size', 'r').read().split()
        return int(columns)

    def Print(self,par,value,value2=None,dest=sys.stdin):
        parout=" - %s %s"%(par,"."*(self.LeftW-len(par)))
        if value2 is None:
            valueOut=value
        else:
            valueOut="%s%s"%(value.ljust(self.WV0),(""" "%s" """%value2).rjust(self.WV1))
        print("%s = %s"%(parout,valueOut), file=dest)
        
    def Print2(self,par,value,helpit,col="white"):
        WidthTerm=self.getWidth()
        Lpar=len(str(par))
        Lval=len(str(value))
        SFill="."*max(self.LeftW-self.Lproto-Lpar-Lval,2)
        WidthHelp=WidthTerm-(self.Lproto+Lpar+Lval+len(SFill))
        Spar="%s"%ModColor.Str(str(par),col=col,Bold=False)
        Sval="%s"%ModColor.Str(str(value),col=col,Bold=False)
        if helpit=="":
            helpit="Help yourself"
        Shelp="%s"%helpit
        if WidthHelp<0:
             print(self.proto%(Spar,SFill,Sval)+Shelp)
             return
        Lhelp=len(str(helpit))
        listStrHelp=list(range(0,Lhelp,WidthHelp))
        
        if listStrHelp[-1]!=Lhelp: listStrHelp.append(Lhelp)
        
        
        
        print(self.proto%(Spar,SFill,Sval)+Shelp[0:WidthHelp])
        for i in range(1,len(listStrHelp)-1):
            parout="%s: %s"%(" "*(self.LeftW-2),Shelp[listStrHelp[i]:listStrHelp[i+1]])
            print(parout)



#!/usr/bin/env python
from __future__ import print_function

from collections import OrderedDict
import matplotlib.patches as mpatches
from pylab import *
from cubical.plots import gainsols

def plot_leakages_cc(D, FS=None, TS=None, ANTS=slice(None), refant=None,
                     plot_diag='ap', plot_offdiag='ri', figtitle=None):
    """Plots leakages from a CubiCal database"""
    sols, have_offdiag, is_complex = gainsols.prepare_sols_dict(D, FS, TS)
    if not have_offdiag:
        plot_offdiag = ''
    if not is_complex:
        plot_diag = plot_diag and 'r'
        plot_offdiag = plot_offdiag and 'r'

    # renormalize
    sols = OrderedDict([(ant, (t, f, d00, d01/d00, d10/d11, d11))
                        for ant, (t, f, d00, d01, d10, d11) in sols.items()])

    # apply new reference antenna
    if refant is not None:
        d01ref = sols[refant][3]
        sols = OrderedDict([(ant, (t, f, d00, d01 - d01ref, d10 + np.conj(d01ref), d11))
                            for ant, (t, f, d00, d01, d10, d11) in sols.items()])

    return gainsols.plot_bandpass(sols, plot_diag=plot_diag, plot_offdiag=plot_offdiag,
                                  gaintype=("Bandpass", "Leakage"), figtitle=figtitle)


def plot_leakages_aips(Daips, FS=None, ANTS=slice(None),
                       plot_diag=None, plot_offdiag='ri', diff=False, figtitle=None):
    """
    Plots AIPS leakage solutions
    """
    Da, all_ants, all_freqs = Daips
    FS = gainsols.get_freq_slice(FS, all_freqs)
    sols = OrderedDict()
    # get valid D solutions
    if isinstance(ANTS, (list, tuple)):
        ANTS = [(ant, all_ants[ant]) for ant in ANTS]
    else:
        ANTS = enumerate(all_ants[ANTS])
    time = np.array([0.])
    freq = all_freqs[FS]
    for iant,ant in ANTS:
        d00 = Da[:, FS, iant, 0, 0]
        d01 = Da[:, FS, iant, 0, 1]
        d10 = Da[:, FS, iant, 1, 0]
        d11 = Da[:, FS, iant, 1, 1]
        sols[ant] = time,freq,d00,d01,d10,d11

    return gainsols.plot_bandpass(sols, plot_diag=plot_diag, plot_offdiag=plot_offdiag,
                         gaintype=("Diff bandpass", "Diff leakage") if diff else ("Bandpass", "Leakage"),
                         figtitle=figtitle)

def read_aips_leakages(filename):
    """
    Reads AIPS leakage solutions from text file.

    Returns:
        leakage_array [(1,NFREQ,NANT,2,2) complex], antenna_names [(NANT) str], frequencies [(NFREQ) float]
    """
    import re, cmath
    from collections import OrderedDict
    allfreqs = set()
    aips_rel_leak = OrderedDict()
    antenna = antdict = None
    for iline, line in enumerate(open(filename).readlines()):
        match = re.match("\s*Antenna #\s+(\d+)\s+ name: (\w+)", line)
        if match:
            antenna, antname = match.groups()
            # print("Found antenna {} ({})".format(antenna, antname))
            antdict = aips_rel_leak.setdefault(antenna, dict(R=OrderedDict(), L=OrderedDict()))
            continue
        fields = line.strip().split()
        if len(fields) == 7:
            try:
                pol = fields[2]
                freq = float(fields[3]) * 1e+6
                amp = float(fields[5])
                phase = float(fields[6])
            except:
                continue
            if antdict is None:
                print("line {}: leakage entry before active antenna entry".format(iline + 1))
                continue
            antdict[pol][freq] = d = amp * cmath.exp(1j * phase * cmath.pi / 180)
            allfreqs.add(freq)
            # print("  {} {} MHz: {}".format(pol, freq, d))
    # filter out missing antennas
    missing = []
    for antenna, antdict in aips_rel_leak.items():
        if (np.array([antdict['R'].values(), antdict['L'].values()]) == 0).all():
            missing.append(antenna)
    if missing:
        print("  missing antennas: {}".format(" ".join(missing)))
    for m in missing:
        del aips_rel_leak[m]
    freqs = np.array(sorted(allfreqs))
    freq_index = {f: i for i, f in enumerate(freqs)}
    ants = list(aips_rel_leak.keys())
    ant_index = {a: i for i, a in enumerate(ants)}
    leakage = np.zeros((1, len(allfreqs), len(ants), 2, 2), np.complex64)
    print("  {} antennas: {}".format(len(ants), " ".join(ants)))
    print("  freqs {} to {} MHz".format(*(freqs[[0, -1]] * 1e-6)))

    for antenna, antdict in aips_rel_leak.items():
        iant = ant_index[antenna]
        for icorr, corr in enumerate('RL'):
            corrdict = antdict[corr]
            ifreqs = [freq_index[f] for f in corrdict.keys()]
            leakage[0, ifreqs, iant, icorr, 1 - icorr] = np.array(corrdict.values())
    return leakage, ants, freqs

def apply_ref_ant(leak, refant, ant_index):
    iref = ant_index.get(refant)
    if iref is None:
        print("unknown reference antenna '{}'".format(refant))
    else:
        refleak = leak[:, :, iref, 0, 1].copy()
        leak[:, :, :, 0, 1] -= refleak[..., np.newaxis]
        leak[:, :, :, 1, 0] += np.conj(refleak)[..., np.newaxis]

def subtract_leakages(leak, antennas, leak0, ant0_index):
    diffleak = leak.copy()
    for iant, ant in enumerate(antennas):
        iant0 = ant0_index.get(ant)
        if iant0 is None:
            print("antenna {} not in second file, difference will be zero'ed".format(ant))
            diffleak[:, :, iant, :, :] = 0
        else:
            diffleak[:, :, iant, :, :] -= leak0[:, :, iant0, :, :]
    return diffleak

def read_cubical_leakages(filename):
    return gainsols.read_cubical_gains(filename, "D")



# CubiCal: a radio interferometric calibration suite
# (c) 2017 Rhodes University & Jonathan S. Kenyon
# http://github.com/ratt-ru/CubiCal
# This code is distributed under the terms of GPLv2, see LICENSE.md for details
"""
Source provider for reading source information from a Tigger lsm.
"""
from six import string_types
import logging
import numpy as np

import montblanc

from montblanc.impl.rime.tensorflow.sources import SourceProvider

import Tigger
import pyrap.tables as pt


class TiggerSourceProvider(SourceProvider):
    """
    A Montblanc-compatible source provider that returns source information from a Tigger sky model.
    """
    def __init__(self, lsm, phase_center, dde_tag='dE'):
        """
        Initialises this source provider.

        Args:
            lsm (str):
                Filename containing the sky model
            phase_center (tuple):
                Observation phase centre, as a RA, Dec tuple
            dde_tag (str or None):
                If set, sources are grouped into multiple directions using the specified tag.

        """

        self.filename = lsm
        self._sm = Tigger.load(lsm)
        self._phase_center = phase_center
        self._use_ddes = bool(dde_tag)
        self._dde_tag = dde_tag

        self._freqs = None

        self._clusters = cluster_sources(self._sm, dde_tag)

        self._cluster_keys = list(self._clusters.keys())
        self._nclus = len(self._cluster_keys)

        self._target_key = 0
        self._target_cluster = self._cluster_keys[self._target_key]
        self._target_reqbeam = "beam"

        self._pnt_sources = \
            self._clusters[self._target_cluster][self._target_reqbeam]["pnt"]
        self._npsrc = len(self._pnt_sources)
        self._gau_sources = \
            self._clusters[self._target_cluster][self._target_reqbeam]["gau"]
        self._ngsrc = len(self._gau_sources)

    def set_direction(self, idir, req_beam="beam"):
        """Sets current direction being simulated.

        Args:
            idir (int):
                Direction number, from 0 to n_dir-1
        """

        self._target_key = idir
        self._target_cluster = self._cluster_keys[self._target_key]
        self._target_reqbeam = req_beam
        self._pnt_sources = \
            self._clusters[self._target_cluster][self._target_reqbeam]["pnt"]
        self._npsrc = len(self._pnt_sources)
        self._gau_sources = \
            self._clusters[self._target_cluster][self._target_reqbeam]["gau"]
        self._ngsrc = len(self._gau_sources)

    def set_frequency(self, frequency):
        """Sets simulated frequencies

        Args:
            frequency (ndarray):
                Array of frequencies associated with a DDID
        """
        self._freqs = frequency

    def name(self):
        """ Returns name of assosciated source provider. This is just the filename, in this case."""
        return self.filename

    def point_lm(self, context):
        """ Returns an lm coordinate array to Montblanc. """

        lm = np.empty(context.shape, context.dtype)

        # Print the array schema
        #Montblanc.log.info(context.array_schema.shape)
        # Print the space of iteration
        #Montblanc.log.info(context.iter_args)

        # Get the extents of the time, baseline and chan dimension
        (lp, up) = context.dim_extents('npsrc')
        assert lm.shape == (up-lp, 2)

        for ind, source in enumerate(self._pnt_sources[lp:up]):

            ra, dec = source.pos.ra, source.pos.dec
            lm[ind,0], lm[ind,1] = ra, dec #radec_to_lm(ra, dec, self._phase_center)

        return lm

    def source_spectrum(self, source, freqs):
        """
        Calculate a spectrum for the source. If source.spectrum attribute is missing,
        returns flat spectrum

        Args:
            source (:obj:`~Tigger.Models.SkyModel.Source`):
                Tigger source
            freq (`numpy.ndarray`):
                array of frequencies of shape `(chan,)`

        Returns:
            `numpy.ndarray`:
                spectrum of shape `(chan,)`
        """
        if hasattr(source, 'spectrum') and source.spectrum is not None:
            rf = getattr(source.spectrum, 'freq0', 1e+9)
            alpha = source.spectrum.spi
            frf = freqs / rf
            if not isinstance(alpha, (list, tuple)):
                alpha = [alpha]
            logfr = np.log10(frf)
            spectrum = frf ** sum([a * np.power(logfr, n) for n, a in enumerate(alpha)])

            ## broadcast into the time dimension.
            return spectrum
        else:
            return np.ones_like(freqs)

    def point_stokes(self, context):
        """ Returns a stokes parameter array to Montblanc. """

        # Get the extents of the time, baseline and chan dimension
        (lp, up), (lt, ut), (lc, uc) = context.dim_extents('npsrc',
                                                           'ntime',
                                                           'nchan')
        # (npsrc, ntime, nchan, 4)
        stokes = np.empty(context.shape, context.dtype)

        f = self._freqs[lc:uc].ravel()

        for ind, source in enumerate(self._pnt_sources[lp:up]):
            spectrum = self.source_spectrum(source, f)[None, :]

            # Multiply flux into the spectrum,
            # broadcasting into the time dimension
            for iS, S in enumerate('IQUV'):
                stokes[ind, :, :, iS] = getattr(source.flux, S, 0.)*spectrum

        return stokes

    def gaussian_lm(self, context):
        """ Returns an lm coordinate array to Montblanc. """

        lm = np.empty(context.shape, context.dtype)

        # Get the extents of the time, baseline and chan dimension
        (lg, ug) = context.dim_extents('ngsrc')

        for ind, source in enumerate(self._gau_sources[lg:ug]):

            ra, dec = source.pos.ra, source.pos.dec
            lm[ind, 0], lm[ind, 1] = ra, dec #radec_to_lm(ra, dec, self._phase_center)

        return lm

    def gaussian_stokes(self, context):
        """ Return a stokes parameter array to Montblanc """

        # Get the extents of the source, time and channel dims
        (lg, ug), (lt, ut), (lc, uc) = context.dim_extents('ngsrc',
                                                           'ntime',
                                                           'nchan')
        # (npsrc, ntime, nchan, 4)
        stokes = np.empty(context.shape, context.dtype)

        f = self._freqs[lc:uc].ravel()

        for ind, source in enumerate(self._gau_sources[lg:ug]):
            spectrum = self.source_spectrum(source, f)[None, :]

            # Multiply flux into the spectrum,
            # broadcasting into the time dimension
            for iS, S in enumerate('IQUV'):
                stokes[ind, :, :, iS] = getattr(source.flux, S, 0.)*spectrum

        return stokes

    def gaussian_shape(self, context):
        """ Returns a Gaussian shape array to Montblanc """

        shapes = np.empty(context.shape, context.dtype)

        (lg, ug) = context.dim_extents('ngsrc')

        for ind, source in enumerate(self._gau_sources[lg:ug]):
            shapes[0, ind] = source.shape.ex * np.sin(source.shape.pa)
            shapes[1, ind] = source.shape.ex * np.cos(source.shape.pa)
            shapes[2, ind] = source.shape.ey/source.shape.ex

        return shapes

    def updated_dimensions(self):
        """ Informs Montblanc of updated dimension sizes. """

        return [('npsrc', self._npsrc),
                ('ngsrc', self._ngsrc)]

    def phase_centre(self, context):
        """ Sets the MB phase direction """
        radec = np.array([self._phase_center[...,-2],
                          self._phase_center[...,-1]], context.dtype)
        return radec

def cluster_sources(sm, dde_tag):
    """
    Groups sources by shape (point/gaussian) and, optionally, direction.

    Args:
        sm (:obj:`~Tigger.Models.SkyModel.SkyModel`):
            SkyModel object containing source information.
        dde_tag (str or None):
            If given, then also group by direction using the given tag.

    Returns:
        dict:
            Dictionary of grouped sources.
    """

    clus = {}

    # cluster sources by value of dde_tag, or if dde_tag is True but not a string,
    # then by their 'cluster' attribute. Within a cluster, split into point sources
    # and Gaussians
    for src in sm.sources:
        dde_cluster = "die"
        if dde_tag:
            tagvalue = src.getTag(dde_tag)
            if tagvalue:
                if isinstance(tagvalue, string_types):
                    dde_cluster = tagvalue
                else:
                    dde_cluster = src.getTag('cluster')

        req_beam = 'nobeam' if src.getTag('nobeam') else 'beam'
        src_type = 'pnt' if src.typecode=='pnt' else 'gau'

        clus.setdefault(dde_cluster, dict(beam=dict(pnt=[], gau=[]),
                                          nobeam=dict(pnt=[], gau=[])))
        clus[dde_cluster][req_beam][src_type].append(src)

    return clus

# def radec_to_lm(ra, dec, phase_center):
#     """
#     DEPRICATED: Montblanc now implements WCS conversions internally

#     Convert right-ascension and declination to direction cosines.

#     Args:
#         ra (float):
#             Right-ascension in radians.
#         dec (float):
#             Declination in radians.
#         phase_center (np.ndarray):
#             The coordinates of the phase center.

#     Returns:
#         tuple:
#             l and m coordinates.

#     """

#     delta_ra = ra - phase_center[...,-2]
#     dec_0 = phase_center[...,-1]

#     l = np.cos(dec)*np.sin(delta_ra)
#     m = np.sin(dec)*np.cos(dec_0) -\
#         np.cos(dec)*np.sin(dec_0)*np.cos(delta_ra)

#     return l, m



# CubiCal: a radio interferometric calibration suite
# (c) 2017 Rhodes University & Jonathan S. Kenyon
# http://github.com/ratt-ru/CubiCal
# This code is distributed under the terms of GPLv2, see LICENSE.md for details
"""
Handles solver statistics.
"""
from __future__ import print_function
from builtins import range
import math
import numpy as np
from future.moves import pickle

from cubical.tools import logger
from cubical.tools import ModColor
from collections import OrderedDict

log = logger.getLogger("stats")

class SolverStats (object):
    """
    SolverStats is a container for various per-chunk statistics collected during the solving 
    process.
    """
    def __init__ (self, obj):
        """
        Initialisation for the SolverStats object.

        Args:
            obj (dict or np.ndarray or file):
                Object from which to initialise the stats object.

        Raises:
            TypeError:
                If obj is not an understood type.
        """

        if type(obj) is np.ndarray:
            self._init_for_chunk(obj)
        elif type(obj) is dict:
            self._concatenate(obj)
        elif type(obj) is file:
            self.load(obj)
        else:
            raise TypeError("can't init SolverStats from object of type %s" % type(obj))

    def _init_for_chunk (self, data):
        """
        Initializes stats for a chunk of data.

        Args:
            data (np.ndarray):
                A CubiCal data block.
        """

        n_tim, n_fre, n_ant, n_ant, n_cor, n_cor = data.shape
        # summary record arrays (per channel-antenna, time-antenna, time-channel)
        dtype = [ ('dv2', 'f8'), ('dr2', 'f8'), ('dv2n', 'i4'), ('dr2n', 'i4'),
                  ('chi2', 'f8'), ('chi2n', 'i4'),
                  ('initchi2', 'f8'), ('initchi2n', 'i4') ]
        self.chanant  = np.rec.array(np.zeros((n_fre, n_ant), dtype))
        self.timeant  = np.rec.array(np.zeros((n_tim, n_ant), dtype))
        self.timechan = np.rec.array(np.zeros((n_tim, n_fre), dtype))

        # other stats: per chunk

        # these are truly per chunk, corresponding to starting and final values
        dtype = [ ('label', 'S32'), ('num_prior_flagged', 'i8'), ('num_data_points', 'i8') ]

        # these have intermediate values as well (e.g. when solving for a chain)
        self._chunk_stats_intermediate_fields = [
                    ('chi2u', 'f8'), ('noise', 'f8'), ('chi2', 'f8'),
                    ('iters', 'i4'),
                    ('num_solutions', 'i4'), ('num_converged', 'i4'), ('num_stalled', 'i4'),
                    ('num_sol_flagged', 'i4'), ('num_mad_flagged', 'i4'),
                    ('frac_converged', 'f8'), ('frac_stalled', 'f8'),
                    ('end_chi2', 'f8') ]

        dtype += self._chunk_stats_intermediate_fields

        self._max_intermediate_fields = 100
        for i in range(self._max_intermediate_fields):
            dtype += [("{}_{}".format(field, i), dt) for field, dt in self._chunk_stats_intermediate_fields]

        self.chunk = np.rec.array(np.zeros((), dtype))

    def save_chunk_stats(self, step):
        for key, _ in self._chunk_stats_intermediate_fields:
            self.chunk["{}_{}".format(key, step)] = self.chunk[key]

    def save(self, filename):
        """
        Pickles contents to file. Better than pickling whole object, as the pickle then only 
        contains standard classes (i.e. don't need CubiCal to read it).

        Args:
            filename (str):
                Name for pickled file.
        """
        with open(filename, 'wb') as pf:
            pickle.dump((self.chanant, self.timeant, self.timechan, self.chunk), pf, 2)

    def load(self, fileobj):
        """
        Loads contents from file object
        """

        self.chanant, self.timeant, self.timechan, self.chunk = pickle.load(fileobj)

    def estimate_noise (self, data, flags, residuals=False):
        """
        Given a data cube with dimensions (n_tim, n_fre, n_ant, n_ant, n_cor, n_cor) and a flag cube
        with dimensions (n_tim, n_fre, n_ant, n_ant), this function estimates the noise in the data 
        by taking the differences between adjacent channels.

        Sum of delta-visibilities squared, and sum of valid terms, is stored in the 'dv2', 'dv2n' or
        'dr2', 'dr2n' summary fields (the latter if residuals=True).

        Args:
            data (np.ndarray):
                A CubiCal data block.
            flags (np.ndarray):
                A the flag block assosciated with data.
            residuals (bool, optional):
                If True, results are stored in the 'dr2' and 'dr2n' fields.

        Returns:
            tuple:
                Noise, inverse_noise_per_antenna_channel_squared, inverse_noise_per_antenna_squared 
                and inverse_noise_per_channel_squared. 
        """

        n_mod, n_tim, n_fre, n_ant, n_ant, n_cor, n_cor = data.shape

        # For only one frequency channel, we can't estimate noise -- return 1.

        if n_fre == 1:
            return 1., np.ones((n_fre, n_ant), np.float32), np.ones(n_ant, np.float32), np.ones(n_fre, np.float32)

        deltaflags = flags!=0
        deltaflags[:, 1:, ...] = deltaflags[:, 1:, ...] | deltaflags[:, :-1, ...]
        deltaflags[:, 0 , ...] = deltaflags[:,   1, ...]

        # Create array for the squared difference between channel-adjacent visibilities.

        deltavis2 = np.zeros((n_tim, n_fre, n_ant, n_ant), np.float32)

        # Square the absolute value of the difference between channel-adjacent visibilities and sum
        # over correlations. Normalize the result by n_cor*n_cor*4. The factor of 4 arises because
        # Var<c1-c2> = Var<c1>+Var<c2> and Var<c>=Var<r>+Var<i>. Thus, the square of the abs 
        # difference between two complex visibilities has contributions from _four_ noise terms.

        # TODO: When fewer than 4 correlations are provided, the normalisation needs to be 
        # different.

        # TODO: Something smart with multiple-model calibration. Currently only the first dataset
        # is used.

        deltavis2[:, 1:, ...]  = np.square(abs(data[0, :, 1:, ...] - data[0, :, :-1, ...])).sum(axis=(-2,-1))
        deltavis2[:, 1:, ...] /= n_cor*n_cor*4
        deltavis2[:, 0 , ...]  = deltavis2[:, 1, ...]

        # The flagged elements are zeroed; we don't have an adequate noise estimate for those
        # channels.

        deltavis2[deltaflags] = 0

        # This flag inversion gives a count of the valid estimates in deltavis2.

        deltaflags = ~deltaflags

        # Sum into the various stats arrays (use dvis2 or dvis2p field, depending on whether 
        # pre- or post-solver noise is being estimated)

        dv2attr = 'dr2'   if residuals else 'dv2'
        dn2attr = 'dr2n'  if residuals else 'dv2n'

        deltavis2_chan_ant = getattr(self.chanant,  dv2attr)[...] = deltavis2.sum(axis=(0, 2))  # sum, per chan, ant
        getattr(self.timeant,  dv2attr)[...] = deltavis2.sum(axis=(1, 2))  # sum, per chan, time
        getattr(self.timechan, dv2attr)[...] = deltavis2.sum(axis=(2, 3))  # sum, per time, chan
        deltanum_chan_ant = getattr(self.chanant,  dn2attr)[...] = deltaflags.sum(axis=(0, 2))  # sum, per chan, ant
        getattr(self.timeant,  dn2attr)[...] = deltaflags.sum(axis=(1, 2))  # sum, per chan, time
        getattr(self.timechan, dn2attr)[...] = deltaflags.sum(axis=(2, 3))  # sum, per time, chan

        # Compute the variance overall, and per antenna, and per channel.

        with np.errstate(divide='ignore', invalid='ignore'):  # ignore division by 0
            inv_var = deltanum_chan_ant.sum() / deltavis2_chan_ant.sum()
            noise_est = math.sqrt(1/inv_var)
            inv_var_antchan =  deltavis2_chan_ant / deltanum_chan_ant
            inv_var_ant  = deltanum_chan_ant.sum(axis=0) / deltavis2_chan_ant.sum(axis=0)
            inv_var_chan = deltanum_chan_ant.sum(axis=1) / deltavis2_chan_ant.sum(axis=1)

        # Isolated but valid channels may end up with no noise estimate at all. Fill one in from noise_est
        validchans = (flags==0).sum(axis=(0,2,3)) != 0
        inv_var_chan[validchans&~np.isfinite(inv_var_chan)] = inv_var

        # Antennas/channels with no data end up with NaNs here, so replace them with 0.
        
        inv_var_antchan[~np.isfinite(inv_var_antchan)] = 0
        inv_var_ant[~np.isfinite(inv_var_ant)] = 0
        inv_var_chan[~np.isfinite(inv_var_chan)] = 0
        
        return noise_est, inv_var_antchan, inv_var_ant, inv_var_chan

    @staticmethod
    def add_records(recarray, recarray2):
        """ Adds two record-type arrays together. """
        
        for field in recarray.dtype.fields.keys():
            recarray[field] += recarray2[field]

    @staticmethod
    def normalize_records(recarray):
        """ Normalizes record-type arrays by dividing each field 'X' by the field 'Xn'. """

        for field in recarray.dtype.fields.keys():
            if field[-1] != 'n':
                nval = recarray[field+'n']
                mask = nval!=0
                recarray[field][mask] /= nval[mask]

        return np.rec.array(recarray)

    def _concatenate(self, stats):
        """
        Concatenates stats from a dictionary (indexed by time_index,freq_index) into a single 
        object.

        Args:
            stats (dict):
                A dictionary of useful solver statistics.
        """
        
        # Get lists of unique time and channel indices occurring in the dict.
        
        times = sorted(set([time for time, _ in stats.keys()]))
        chans = sorted(set([chan for _, chan in stats.keys()]))

        # Concatenate and add up cumulative stats.
        
        self.chanant = np.concatenate([stats[times[0], chan].chanant for chan in chans], axis=0)

        for time in times[1:]:
            self.add_records(
                self.chanant, np.concatenate([stats[time, chan].chanant for chan in chans], axis=0))

        self.timeant = np.concatenate([stats[time, chans[0]].timeant for time in times], axis=0)

        for chan in chans[1:]:
            self.add_records(self.timeant, 
                np.concatenate([stats[time, chan].timeant for time in times], axis=0))

        self.timechan = \
            np.concatenate([np.concatenate([stats[time, chan].timechan for time in times], axis=0)
                for chan in chans], axis=1)

        # Note that for some reason np.concatenate of record arrays produces structured arrays 
        # instead of nd.recarrays - normalize_records() will convert them back.

        # Normalize by number of values.
        
        self.chanant = self.normalize_records(self.chanant)
        self.timeant = self.normalize_records(self.timeant)
        self.timechan = self.normalize_records(self.timechan)

        # Make 2D array of per-chunk values.

        self.chunk = np.rec.array([[stats[time, chan].chunk for chan in chans] for time in times],
                                  dtype=stats[times[0], chans[0]].chunk.dtype)

    def get_notrivial_chunk_statfields(self):
        """Returns list of interesting (i.e. non-0) stat fields"""
        return [field for field in self.chunk.dtype.names if field != "label" and (self.chunk[field]!=0).any()]

    def format_chunk_stats(self, format_string, ncol=8, threshold=None):
        """
        :param format: format string applied to each record
        :param maxcol: maximum number of columns to allocate
        :return:
        """
        nt, nf = self.chunk.shape
        nt_per_col = 1
        nf_per_col = None
        if nf < ncol:
            nt_per_col = ncol//nf
        else:
            nf_per_col = ncol
        # convert stats to list of columns
        output_rows  = [[("", False)]]
        for itime in range(nt):
            # start new line every NT_PER_COL-th time chunk
            if itime%nt_per_col == 0:
                output_rows.append([])
            for ifreq in range(nf):
                # start new line every NF_PER_COL-th freq chunk, if frequencies span lines
                if nf_per_col is not None and output_rows[-1] and ifreq%nf_per_col == 0:
                    output_rows.append([])
                statrec = self.chunk[itime, ifreq]
                statrec_dict = {field:statrec[field] for field in self.chunk.dtype.fields}
                # new line: prepend chunk label
                label = statrec.label.decode() if hasattr(statrec.label, 'decode') else statrec.label
                if not output_rows[-1]:
                    output_rows[-1].append((label, False))
                # put it in header as well
                if len(output_rows) == 2:
                    output_rows[0].append((label, False))
                # check for threshold
                warn = False
                if threshold is not None:
                    for field, value in threshold:
                        if statrec[field] > value:
                            warn = True
                text = format_string.format(**statrec_dict)
                output_rows[-1].append((text, warn))

        # now work out column widths and format
        ncol = max([len(row) for row in output_rows])
        colwidths = [max([len(row[icol][0]) for row in output_rows if icol<len(row)]) for icol in range(ncol)]
        colformat = ["{{:{}}}  ".format(w) for w in colwidths]

        output_rows = [[(colformat[icol].format(col), warn) for icol, (col, warn) in enumerate(row)] for row in output_rows]

        return ["".join([(ModColor.Str(col, 'red') if warn else col) for col, warn in row]) for row in output_rows]

    def apply_flagcube(self, flag3):
        """
        Applies additional flag cube to statistics. Basically, if something is flagged in the 
        output based on chi-sq or other criteria, we want to remove it from the stats.

        Args:
            flag3 (np.ndarray):
                An (n_times, n_ddid, n_chan) block of flags.
        """

        # Out stats are n_tim, n_fre -- reform cube.

        n_tim, n_ddid, n_fre = flag3.shape
        flag3 = flag3.reshape((n_tim, n_ddid*n_fre))

        FIELDS = list(self.timeant.dtype.fields.keys())

        flagged_times = flag3.all(axis=1)
        flagged_chans = flag3.all(axis=0)

        print("adjusting statistics based on output flags", file=log)
        print("  {:.2%} of all timeslots are flagged".format(
                                        flagged_times.sum()/float(flagged_times.size)), file=log)
        print("  {:.2%} of all channels are flagged".format(
                                        flagged_chans.sum()/float(flagged_chans.size)), file=log)

        for field in FIELDS:
            self.chanant[field][flagged_chans, :] = 0
            self.timeant[field][flagged_times, :] = 0
            self.timechan[field][flagged_times, :] = 0
            self.timechan[field][:, flagged_chans] = 0







# CubiCal: a radio interferometric calibration suite
# (c) 2017 Rhodes University & Jonathan S. Kenyon
# http://github.com/ratt-ru/CubiCal
# This code is distributed under the terms of GPLv2, see LICENSE.md for details
import numpy as np
import traceback
import os.path
import importlib

# number of OMP threads to run, for OMP-related kernels
num_omp_threads = 0

use_parallel = False
use_cache = False

def allocate_reordered_array(shape, dtype, order, zeros=False):
    """
    Allocates an array with its dimensions laid out in a non-default order.
    
    Args:
        shape: shape of resulting array
        dtype: type
        order: order of dims in memory. E.g. [2,0,1] means the last dim changes slowest,
            then dim 0, then dim 1. If N=len(shape) < M=len(order), then the dims referred to
            by 'shape' are assumed to refer to the *trailing* N dimensions in 'order'.
            This allows 'order' to describe the layout of e.g. a (dir,model,time,freq,...) array,
            while the function can still be used to allocate a (time,freq,...) array.
        zeros: if True, array will be zeroed, else uninitialized. 

    Returns:
        View onto allocated array, of the given shape.
    """
    extra_dims = len(order) - len(shape)
    assert(extra_dims>=0)
    # take care of extra leading dimensions
    if extra_dims:
        slicer = tuple([0]*extra_dims)
        shape = [1]*extra_dims + list(shape)
    else:
        slicer = None

    _intrinsic_shape = [shape[i] for i in order]
    array = (np.zeros if zeros else np.empty)(_intrinsic_shape, dtype=dtype)
    array = array.transpose(np.argsort(order))

    if extra_dims:
        array = array[slicer]

    return array

# _omp_import = [False]

def import_kernel(name):
    """ Imports named kernel and returns it. """

    return importlib.import_module("cubical.kernels." + name)








# CubiCal: a radio interferometric calibration suite
# (c) 2017 Rhodes University & Jonathan S. Kenyon
# http://github.com/ratt-ru/CubiCal
# This code is distributed under the terms of GPLv2, see LICENSE.md for details
"""
Cython kernels for the robust 2x2 complex gain machine. Functions require output 
arrays to be provided. Common dimensions of arrays are:

+----------------+------+
| Dimension      | Size |
+================+======+
| Direction      |   d  |
+----------------+------+
| Model          |   m  |
+----------------+------+
| Time           |   t  |
+----------------+------+
| Time Intervals |   ti |
+----------------+------+
| Frequency      |   f  |
+----------------+------+
| Freq Intervals |   fi |
+----------------+------+
| Antenna        |   a  |
+----------------+------+
| Correlation    |   c  |
+----------------+------+

"""
from builtins import range

import numpy as np
from numba import jit, prange

import cubical.kernels
from cubical.kernels import generics
from cubical.kernels import full_complex

use_parallel = True if cubical.kernels.num_omp_threads > 1 else False
use_cache = cubical.kernels.use_cache

# defines memory layout of model-like arrays (axis layout is NDxNMxNTxNFxNAxNAxNCxNC)
_model_axis_layout = [4,5,1,2,3,0,6,7]    # layout is AAMTFD

# defines memory layout of gain-like arrays  (axis layout is NDxNTxNFxNAxNCxNC)
_gain_axis_layout = [3,1,2,0,4,5]   # layout is ATFD

# defines memory layout of flag-like arrays  (axis layout is NTxNFxNAxNA)
_flag_axis_layout = [2,3,0,1]      # layout is AATF

def allocate_vis_array(shape, dtype, zeros=False):
    """Allocates a visibility array of the desired shape, laid out in preferred order"""
    return cubical.kernels.allocate_reordered_array(shape, dtype, _model_axis_layout, zeros=zeros)

def allocate_gain_array(shape, dtype, zeros=False):
    """Allocates a gain array of the desired shape, laid out in preferred order"""
    return cubical.kernels.allocate_reordered_array(shape, dtype, _gain_axis_layout, zeros=zeros)

def allocate_flag_array(shape, dtype, zeros=False):
    """Allocates a flag array of the desired shape, laid out in preferred order"""
    return cubical.kernels.allocate_reordered_array(shape, dtype, _flag_axis_layout, zeros=zeros)

allocate_param_array = allocate_gain_array

# compute_residual is identical to the general full complex case.
compute_residual = full_complex.compute_residual

# compute_jh is identical to the general full complex case.
compute_jh = full_complex.compute_jh

# compute_update is identical to the general full complex case.
compute_update = full_complex.compute_update

# compute_corrected is identical to the general full complex case.
compute_corrected = full_complex.compute_corrected

# apply_gains is identical to the general full complex case.
apply_gains = full_complex.apply_gains

# right_multiply_gains is identical to the general full complex case.
right_multiply_gains = full_complex.right_multiply_gains

# Map the J^H.J inversion method to a generic inversion.
compute_jhjinv = generics.compute_2x2_inverse

# Map inversion to generic 2x2 inverse.
invert_gains = generics.compute_2x2_inverse

@jit(nopython=True, fastmath=True, parallel=use_parallel, cache=use_cache, nogil=True)
def compute_jhwr(jh, r, w, jhwr, t_int, f_int):
    """
    Given J\ :sup:`H` and the residual (or observed data, in special cases), computes J\ :sup:`H`\R.
    J\ :sup:`H`\R is computed over intervals.

    Args:
        jh (np.complex64 or np.complex128):
            Typed memoryview of J\ :sup:`H` array with dimensions (d, m, t, f, a, a, c, c).
        r (np.complex64 or np.complex128):
            Typed memoryview of residual array with dimensions (m, t, f, a, a, c, c).
        w (np.complex64 or np.complex128):
            Typed memoryview of residual array with dimensions (m, t, f, a, a, 1).
        jhwr (np.complex64 or np.complex128):
            Typed memoryview of J\ :sup:`H`\R array with dimensions (d, ti, fi, a, c, c).
        t_int (int):
            Number of time slots per solution interval.
        f_int (int):
            Number of frequencies per solution interval.
    """
    n_dir = jh.shape[0]
    n_mod = jh.shape[1]
    n_tim = jh.shape[2]
    n_fre = jh.shape[3]
    n_ant = jh.shape[4]

    all_bls = np.array([[i,j] for i in range(n_ant) for j in range(n_ant) if i!=j], dtype=np.int32)
    n_bl = all_bls.shape[0]

    broadcast_times = np.array([t//t_int for t in range(n_tim)])
    broadcast_freqs = np.array([f//f_int for f in range(n_fre)])

    for ibl in prange(n_bl):
        aa, ab = all_bls[ibl][0], all_bls[ibl][1]
        for i in range(n_mod):
            for t in range(n_tim):
                bt = broadcast_times[t]
                for f in range(n_fre):
                    bf = broadcast_freqs[f]
                    for d in range(n_dir):

                        r00 = r[i,t,f,aa,ab,0,0]
                        r01 = r[i,t,f,aa,ab,0,1]
                        r10 = r[i,t,f,aa,ab,1,0]
                        r11 = r[i,t,f,aa,ab,1,1]

                        jhh00 = jh[d,i,t,f,ab,aa,0,0]
                        jhh01 = jh[d,i,t,f,ab,aa,0,1]
                        jhh10 = jh[d,i,t,f,ab,aa,1,0]
                        jhh11 = jh[d,i,t,f,ab,aa,1,1]

                        w0 = w[i,t,f,aa,ab,0]

                        jhwr[d,bt,bf,aa,0,0] += (r00*jhh00 + r01*jhh10)*w0 
                        jhwr[d,bt,bf,aa,0,1] += (r00*jhh01 + r01*jhh11)*w0 
                        jhwr[d,bt,bf,aa,1,0] += (r10*jhh00 + r11*jhh10)*w0 
                        jhwr[d,bt,bf,aa,1,1] += (r10*jhh01 + r11*jhh11)*w0 

@jit(nopython=True, fastmath=True, parallel=use_parallel, cache=use_cache, nogil=True)
def compute_jhwj(jh, w, jhwj, t_int, f_int):
    """
    Given J\ :sup:`H` ,computes the diagonal entries of J\ :sup:`H`\J. J\ :sup:`H`\J is computed
    over intervals. This is an approximation of the Hessian.

    Args:
        jh (np.complex64 or np.complex128):
            Typed memoryview of J\ :sup:`H` array with dimensions (d, m, t, f, a, a, c, c).
        w (np.complex64 or np.complex128):
            Typed memoryview of residual array with dimensions (m, t, f, a, a, 1).
        jhj (np.complex64 or np.complex128):
            Typed memoryview of J\ :sup:`H`\J array with dimensions (d, ti, fi, a, c, c).
        t_int (int):
            Number of time slots per solution interval.
        f_int (int):
            Number of frequencies per solution interval.
    """
    n_dir = jh.shape[0]
    n_mod = jh.shape[1]
    n_tim = jh.shape[2]
    n_fre = jh.shape[3]
    n_ant = jh.shape[4]

    all_bls = np.array([[i,j] for i in range(n_ant) for j in range(n_ant) if i!=j], dtype=np.int32)
    n_bl = all_bls.shape[0]

    broadcast_times = np.array([t//t_int for t in range(n_tim)])
    broadcast_freqs = np.array([f//f_int for f in range(n_fre)])

    for ibl in prange(n_bl):
        aa, ab = all_bls[ibl][0], all_bls[ibl][1]
        for i in range(n_mod):
            for t in range(n_tim):
                bt = broadcast_times[t]
                for f in range(n_fre):
                    bf = broadcast_freqs[f]
                    for d in range(n_dir):

                        j00 = jh[d,i,t,f,ab,aa,0,0]
                        j01 = jh[d,i,t,f,ab,aa,0,1]
                        j10 = jh[d,i,t,f,ab,aa,1,0] 
                        j11 = jh[d,i,t,f,ab,aa,1,1]

                        jh00 = jh[d,i,t,f,ab,aa,0,0].conjugate()
                        jh01 = jh[d,i,t,f,ab,aa,1,0].conjugate()
                        jh10 = jh[d,i,t,f,ab,aa,0,1].conjugate() 
                        jh11 = jh[d,i,t,f,ab,aa,1,1].conjugate()

                        w0 = w[i,t,f,aa,ab,0]

                        jhwj[d,bt,bf,aa,0,0] += (jh00*j00 + jh01*j10)*w0
                        jhwj[d,bt,bf,aa,0,1] += (jh00*j01 + jh01*j11)*w0
                        jhwj[d,bt,bf,aa,1,0] += (jh10*j00 + jh11*j10)*w0
                        jhwj[d,bt,bf,aa,1,1] += (jh10*j01 + jh11*j11)*w0

@jit(nopython=True, fastmath=True, parallel=use_parallel, cache=use_cache, nogil=True)
def compute_weights(r, ic, w, v, npol):
    """
    This function updates the weights, using the expression: 
        w[i] = (v+2*npol)/(v + 2*r[i].T.cov.r[i])

    Args:
        r (np.complex64 or np.complex128):
            Typed memoryview of residual array with dimensions (m, t, f, a, a, c, c).
        ic (np.complex64 or np.complex128):
            Typed memoryview of inverse covariance array with dimensions (4,4).
        w (np.complex64 or np.complex128):
            Typed memoryview of weight array with dimensions (m, t, f, a, a, 1).
        v (float):
            Degrees of freedom of the t-distribution.
        npol (float):
            Number of polarizations (correlations) in use.            
    """
    n_mod = r.shape[0]
    n_tim = r.shape[1]
    n_fre = r.shape[2]
    n_ant = r.shape[3]

    all_bls = np.array([[i,j] for i in range(n_ant) for j in range(n_ant) if i!=j], dtype=np.int32)
    n_bl = all_bls.shape[0]
    
    for ibl in prange(n_bl):
        aa, ab = all_bls[ibl][0], all_bls[ibl][1]
        for i in range(n_mod):
            for t in range(n_tim):
                for f in range(n_fre):

                    r00 = r[i,t,f,aa,ab,0,0]
                    r01 = r[i,t,f,aa,ab,0,1]
                    r10 = r[i,t,f,aa,ab,1,0]
                    r11 = r[i,t,f,aa,ab,1,1]

                    rc00 = r[i,t,f,aa,ab,0,0].conjugate()
                    rc01 = r[i,t,f,aa,ab,0,1].conjugate()
                    rc10 = r[i,t,f,aa,ab,1,0].conjugate()
                    rc11 = r[i,t,f,aa,ab,1,1].conjugate()

                    denom = (rc00*ic[0,0] + rc01*ic[1,0] + rc10*ic[2,0] + rc11*ic[3,0])*r00 + \
                            (rc00*ic[0,1] + rc01*ic[1,1] + rc10*ic[2,1] + rc11*ic[3,1])*r01 + \
                            (rc00*ic[0,2] + rc01*ic[1,2] + rc10*ic[2,2] + rc11*ic[3,2])*r10 + \
                            (rc00*ic[0,3] + rc01*ic[1,3] + rc10*ic[2,3] + rc11*ic[3,3])*r11

                    w[i,t,f,aa,ab,0] = (v + npol)/(v + denom.real) # using LB derivation     

@jit(nopython=True, fastmath=True, parallel=use_parallel, cache=use_cache, nogil=True)
def compute_cov(r, ic, w):
    """
    This function computes the un-normlaised weighted covariance matrix of 
    the visibilities using the expression: 
        cov = r.conj()*w.r
    
    Args:
        r (np.complex64 or np.complex128):
            Typed memoryview of residual array with dimensions (m, t, f, a, a, c, c).
        ic (np.complex64 or np.complex128):
            Typed memoryview of weighted inverse covariance array with dimensions (4,4).
        w (np.complex64 or np.complex128):
            Typed memoryview of weight array with dimensions (m, t, f, a, a, 1).         
    """    
    n_mod = r.shape[0]
    n_tim = r.shape[1]
    n_fre = r.shape[2]
    n_ant = r.shape[3]

    bls = np.array([[i,j] for i in range(n_ant) for j in range(i+1, n_ant)], dtype=np.int32)
    n_bl = bls.shape[0]
    
    for ibl in prange(n_bl):
        aa, ab = bls[ibl][0], bls[ibl][1]  
        for i in range(n_mod):
            for t in range(n_tim):
                for f in range(n_fre):

                    r00 = r[i,t,f,aa,ab,0,0]
                    r01 = r[i,t,f,aa,ab,0,1]
                    r10 = r[i,t,f,aa,ab,1,0]
                    r11 = r[i,t,f,aa,ab,1,1]

                    rc00 = r[i,t,f,aa,ab,0,0].conjugate()
                    rc01 = r[i,t,f,aa,ab,0,1].conjugate()
                    rc10 = r[i,t,f,aa,ab,1,0].conjugate()
                    rc11 = r[i,t,f,aa,ab,1,1].conjugate()

                    w0 = w[i,t,f,aa,ab,0].real

                    w0r00 = w0*r00
                    w0r01 = w0*r01
                    w0r10 = w0*r10
                    w0r11 = w0*r11

                    ic[0,0] += rc00*w0r00
                    ic[0,1] += rc00*w0r01
                    ic[0,2] += rc00*w0r10
                    ic[0,3] += rc00*w0r11

                    ic[1,0] += rc01*w0r00
                    ic[1,1] += rc01*w0r01
                    ic[1,2] += rc01*w0r10
                    ic[1,3] += rc01*w0r11

                    ic[2,0] += rc10*w0r00
                    ic[2,1] += rc10*w0r01
                    ic[2,2] += rc10*w0r10
                    ic[2,3] += rc10*w0r11

                    ic[3,0] += rc11*w0r00
                    ic[3,1] += rc11*w0r01
                    ic[3,2] += rc11*w0r10
                    ic[3,3] += rc11*w0r11


# CubiCal: a radio interferometric calibration suite
# (c) 2017 Rhodes University & Jonathan S. Kenyon
# http://github.com/ratt-ru/CubiCal
# This code is distributed under the terms of GPLv2, see LICENSE.md for details
"""
Handles parameter databases which can contain solutions and other relevant values. 
"""
import numpy as np
from cubical.tools import logger
log = logger.getLogger("param_db")

#from database.pickled_db import PickledDatabase
from .database.casa_db_adaptor import casa_db_adaptor
def create(filename, metadata={}, backup=True):
    """
    Creates a new parameter database.
    
    Args:
        filename (str): 
            Name of file to save DB to.
        metadata (dict, optional): 
            Optional dictionary of metadata.
        backup (bool, optional):
            If True, and an old database with the same filename exists, make a backup.

    Returns:
        :obj:`~cubical.param_db.PickledDatabase`:
            A resulting parameter database.
    """

    db = casa_db_adaptor()
    db._create(filename, metadata, backup)
    
    return db

def load(filename):
    """
    Loads a parameter database

    Args:
        filename (str): 
            Name of file to load DB from.

    Returns:
        :obj:`~cubical.param_db.PickledDatabase`:
            A resulting parameter database.
    """

    db = casa_db_adaptor()
    db._load(filename)
    
    return db



if __name__ == "__main__":
    log.verbosity(2)
    print("Creating test DB")
    db = create("test.db")
    db.define_param("G", np.float64,
                    ["ant", "time", "freq", "corr"], interpolation_axes=["time", "freq"])
    db.define_param("B", np.float64,
                    ["ant", "time", "freq", "corr"], interpolation_axes=["time", "freq"])
    for i0,i1 in (0,2),(4,6),(7,9):
        arr = np.full((3,i1-i0,1,2), i0, float)
        db.add_chunk("G", arr, grid=dict(time=np.arange(i0,i1)))
        arr = np.full((3,1,i1-i0,2), i0, float)
        db.add_chunk("B", arr, grid=dict(freq=np.arange(i0,i1)))
    db.close()

    print("Loading test DB")
    db = load("test.db")
    print(db.names())
    G = db['G']
    B = db['B']
    print("G", db["G"].axis_labels, db["G"].shape)
    print("B", db["B"].axis_labels, db["B"].shape)
    print("G", G.get_slice(ant=0,corr=0))
    print("B", G.get_slice(ant=0,corr=0))
    print("Gint", G.reinterpolate(time=np.arange(0,10,.5),freq=np.arange(0,10,1.5)))




# CubiCal: a radio interferometric calibration suite
# (c) 2017 Rhodes University & Jonathan S. Kenyon
# http://github.com/ratt-ru/CubiCal
# This code is distributed under the terms of GPLv2, see LICENSE.md for details
"""
Handles the flagging of data. 
"""

# This is to keep matplotlib from falling over when no DISPLAY is set (which it otherwise does, 
# even if one is only trying to save figures to .png.
from __future__ import print_function
from builtins import range
from past.builtins import cmp
from functools import cmp_to_key
import numpy as np
import re

from cubical.tools import logger, ModColor
import cubical.plots as plots
log = logger.getLogger("flagging")
from collections import OrderedDict

class FL(object):
    """ Namespace for flag bits. """

    dtype = np.uint16   # dtype used for flag arrays

    PRIOR    = dtype(1<<0)    # prior flags (i.e. from MS)
    MISSING  = dtype(1<<1)    # missing data
    INVALID  = dtype(1<<2)    # invalid data (zero, inf, nan)
    ILLCOND  = dtype(1<<3)    # solution ill conditioned - bad inverse
    DIVERGE  = dtype(1<<4)    # divergence
    NOSOL    = dtype(1<<5)    # missing solution (i.e. in load-from or xfer-from)
    GOOB     = dtype(1<<6)    # gain solution out of bounds
    BOOM     = dtype(1<<7)    # gain solution exploded (i.e. went to inf/nan)
    GNULL    = dtype(1<<8)    # gain solution gone to zero.
    LOWSNR   = dtype(1<<9)    # prior SNR too low for gain solution
    GVAR     = dtype(1<<10)   # posterior variance too high for gain solution
    INVMODEL = dtype(1<<11)   # invalid model (zero, inf, nan)
    INVWGHT  = dtype(1<<12)   # invalid weight (inf or nan)
    NULLWGHT = dtype(1<<13)   # null weight

    MAD      = dtype(1<<14)   # residual exceeds MAD-based threshold
    SKIPSOL  = dtype(1<<15)   # omit this data point from the solver (not a true flag!)

    @staticmethod
    def categories():
        """ Returns dict of all possible flag categories. """

        return OrderedDict([(attr, value) for attr, value in FL.__dict__.items()
                            if attr[0] != "_" and type(value) is FL.dtype])

class Flagsets (object):
    """ Manage an measurement set's flagsets. Pasted from Cattery.Meow.MSUtils. """

    def __init__ (self,ms):
        """
        Initialises a Flagsets object for the measurement set.

        Args:
            ms (:obj:`~casacore.tables.table.table`):
                A table object belonging to the measurement set.
        """

        self.ms = ms
        if not 'BITFLAG' in ms.colnames():
            self.order = None
            self.bits = {}
        else:
            kws = ms.colkeywordnames('BITFLAG')
            self.bits  = {}
            # scan FLAGSET_xxx keywords and populate name->bitmask mappings
            for kw in kws:
                match = re.match('^FLAGSET_(.*)$',kw)
                if match:
                    name = match.group(1)
                    bit = ms.getcolkeyword('BITFLAG',kw)
                    if isinstance(bit,int):
                        self.bits[name] = bit
                    else:
                        print("Warning: unexpected type (%s) for %s keyword of BITFLAG column," \
                                " ignoring"%(type(bit),kw))
            # have we found any FLAGSET_ specs?
            if self.bits:
                order = 'FLAGSETS' in kws and ms.getcolkeyword('BITFLAG','FLAGSETS')
                if isinstance(order,str):
                    order = order.split(',')
                else:
                    print("Warning: unexpected type (%s) for FLAGSETS keyword of BITFLAG column," \
                                " ignoring"%type(order))
                    order = []
                # form up "natural" order by comparing bitmasks
                bitwise_order = list(self.bits.keys())
                bitwise_order.sort(key=cmp_to_key(lambda a,b:cmp(self.bits[a],self.bits[b])))
                # if an order is specified, make sure it is actually valid,
                # and add any elements from bitwise_order that are not present
                self.order = [ fs for fs in order if fs in self.bits ] + \
                             [ fs for fs in bitwise_order if fs not in order ]
                # if order was fixed compared to what was in MS, write back to MS
                if ms.iswritable() and self.order != order:
                    ms._putkeyword('BITFLAG','FLAGSETS',-1,False,','.join(self.order))
                    ms.flush()
            # else if no flagsets found, try the old-style NAMES keyword
            elif 'NAMES' in kws:
                names = ms.getcolkeyword('BITFLAG','NAMES')
                if isinstance(names,(list,tuple)):
                    self.order = list(map(str,names))
                    bit = 1
                    for name in self.order:
                        self.bits[name] = bit
                        bit <<= 1
                    if ms.iswritable():
                        ms._putkeyword('BITFLAG','FLAGSETS',-1,False,','.join(self.order))
                        for name,bit in self.bits.items():
                            ms._putkeyword('BITFLAG','FLAGSET_%s'%name,-1,False,bit)
                        ms.flush()
            else:
                self.order = []

    def names (self):
        """
        Convenience function for determining active flagsets.

        Returns:
            list or None:
                A list of flagset names, in the order in which they were created or None if BITFLAG 
                column is missing (so flagsets are unavailable.)
        """

        return self.order

    def flagmask (self,name,create=False):
        """
        Flagmask corresponding to named flagset.

        Args:
            name (str):
                Name of flagset.
            create (bool, optional): 
                If True and flagset is missing, creates named flagset, else raises exception.
    
        Raises:
            TypeError:
                If the MS does not contain a BITFLAG column.
            ValueError:
                If the named flagset is not found and create is False. 
            ValueError:
                If there are too many flagsets to create a new one.
        """

        # Cludge for python2/3 interoperability.
        name = str(name)

        # lookup flagbit, return if found
        if self.order is None:
            raise TypeError("MS does not contain a BITFLAG column. Please run the addbitflagcol" \
                                                                        " utility on this MS.")
        bit = self.bits.get(name,None)
        if bit is not None:
            return bit
        # raise exception if not allowed to create a new one
        if not create:
            raise ValueError("Flagset '%s' not found"%name)
        # find empty bit
        for bitnum in range(32):
            bit = 1<<bitnum
            if bit not in list(self.bits.values()):
                self.order.append(name)
                self.bits[name] = bit
                self.ms._putkeyword('BITFLAG','FLAGSETS',-1,False,','.join(self.order))
                self.ms._putkeyword('BITFLAG','FLAGSET_%s'%name,-1,False,bit)
                self.ms.flush()
                return bit
        # no free bit found, bummer
        raise ValueError("Too many flagsets in MS, cannot create another one")

    def remove_flagset (self, *fsnames):
        """
        Removes the named flagset(s). 

        Args:
            fsnames (tuple):
                Names of flagsets to be removed.

        Returns:
            int:
                Flagmask corresponding to the removed flagsets.
        """
        
        # lookup all flagsets, raise error if any not found
        if self.bits is None:
            raise TypeError("MS does not contain a BITFLAG column, cannot use flagsets")
        removing = []
        for fs in fsnames:
            bit = self.bits.get(fs,None)
            if bit is None:
                raise ValueError("Flagset '%s' not found"%fs)
            removing.append((fs,bit))
        if not removing:
            return
        # remove items, form up mask of bitflags to be cleared
        mask = 0
        for name,bit in removing:
            mask |= bit
            del self.bits[name]
            del self.order[self.order.index(name)]
            self.ms.removecolkeyword('BITFLAG','FLAGSET_%s'%name)
        # write new list of bitflags
        self.ms._putkeyword('BITFLAG','FLAGSETS',-1,False,','.join(self.order))
        self.ms.flush()

        return mask

def flag_chisq (st, GD, basename, nddid):
    """
    Flags timeslots and channels based on accumulated chi-squared statistics.

    Args:
        st (:obj:`~cubical.statistics.SolverStats`):
            Object containing solver statistics.
        GD (dict):
            Dictionary of global options.
        basename (str):
            Base name for output plots.
        nddid (int):
            Number of data descriptor identifiers.

    Returns:
        np.ndarray:
            Flag cube of shape (n_times, n_ddids, n_chans).
    """

    chi2 = np.ma.masked_array(st.timechan.chi2, st.timechan.chi2==0)
    total = (~chi2.mask).sum()
    if not total:
        print(ModColor.Str("no valid solutions anywhere: skipping post-solution flagging."), file=log)
        return None

    chi2n = st.timechan.chi2n
    chi2n = np.ma.masked_array(chi2n, chi2n == 0)

    median = np.ma.median(chi2)
    median_np = np.ma.median(chi2n)
    print("median chi2 value is {:.3} from {} valid t/f slots".format(median, total), file=log)
    print("median count per slot is {}".format(median_np), file=log)

    chi_median_thresh = GD["postmortem"]["tf-chisq-median"]
    np_median_thresh  = GD["postmortem"]["tf-np-median"]
    time_density      = GD["postmortem"]["time-density"]
    chan_density      = GD["postmortem"]["chan-density"]
    ddid_density      = GD["postmortem"]["ddid-density"]

    make_plots = GD["out"]["plots"]
    show_plots = make_plots == "show"

    if make_plots:
        import pylab
        pylab.figure(figsize=(32, 10))
        pylab.subplot(161)
        if chi2.count():
            pylab.imshow(chi2, vmin=0, vmax=5 * median)
        pylab.title("$\chi^2$")
        pylab.colorbar()
        pylab.subplot(162)
        if chi2n.count():
            pylab.imshow(chi2n)
        pylab.title("counts")
        pylab.colorbar()

    flag = (chi2 > chi_median_thresh * median)
    chi2[flag] = np.ma.masked
    nflag = flag.sum()
    print("{} slots ({:.2%}) flagged on chi2 > {}*median".format(nflag, nflag/float(total), 
                                                                                chi_median_thresh), file=log)

    if make_plots:
        pylab.subplot(163)
        if chi2.count():
            pylab.imshow(chi2)
        pylab.title("$\chi^2$ median flagged")
        pylab.colorbar()

    flag2 = (chi2n < np_median_thresh * median_np)
    n_new = (flag2&~flag).sum()
    print("{} more slots ({:.2%}) flagged on counts < {}*median".format(n_new, 
                                                            n_new/float(total), np_median_thresh), file=log)
    flag |= flag2

    chi2[flag] = np.ma.masked
    if make_plots:
        pylab.subplot(164)
        if chi2.count():
            pylab.imshow(chi2)
        pylab.title("counts flagged")
        pylab.colorbar()

    nt, nf = flag.shape

    # flag channels with overdense flagging
    freqcount = flag.sum(axis=0)
    freqflags = freqcount > nt * chan_density
    n_new = (freqflags&~(freqcount==nt)).sum()
    print("{} more channels flagged on density > {}".format(n_new, chan_density), file=log)

    # flag timeslots with overdense flagging
    timecount = flag.sum(axis=1)
    timeflags = timecount > nf * time_density
    n_new = (timeflags&~(timecount==nf)).sum()
    print("{} more timeslots flagged on density > {}".format(n_new, time_density), file=log)

    flag = flag | freqflags[np.newaxis,:] | timeflags[:,np.newaxis]
    chi2[flag] = np.ma.masked
    if make_plots:
        pylab.subplot(165)
        if chi2.count():
            pylab.imshow(chi2)
        pylab.title("overdense flagged")
        pylab.colorbar()

    # reshape flag array into time, ddid, channel
    flag3 = flag.reshape((nt, nddid, nf // nddid))

    # flag entire DDIDs with overdense flagging
    maxcount = nt*nf/nddid
    ddidcounts = flag3.sum(axis=(0, 2))
    ddidflags = ddidcounts > maxcount * ddid_density
    n_new = (ddidflags&~(ddidcounts==maxcount)).sum()
    print("{} more ddids flagged on density > {}".format(n_new, ddid_density), file=log)

    flag3 |= ddidflags[np.newaxis, :, np.newaxis]
    chi2[flag] = np.ma.masked
    if make_plots:
        pylab.subplot(166)
        pylab.title("overdense DDID")
        if chi2.count():
            pylab.imshow(chi2)
        pylab.colorbar()
        filename = basename+".chiflag.png"
        pylab.savefig(filename, DPI=plots.DPI)
        print("saved chi-sq flagging plot to "+filename, file=log)
        if show_plots:
            pylab.show()

    return flag3



import numpy as np
import scipy.stats as sstats
import scipy.signal as ssig
import scipy.spatial as spat
import copy
from cubical.degridder.geometry import (BoundingConvexHull,
                                        BoundingBox,
                                        BoundingBoxFactory)
DEBUG=False
def test_hull_construction():
    # test case 1
    vals = np.array([[50, 60], [20, 40], [-74, 50], [-95, +10], [20, 60]])
    bh = BoundingConvexHull(vals)
    mask = bh.mask 
    assert mask.shape == (np.max(vals[:, 1]) - np.min(vals[:, 1]) + 1, np.max(vals[:, 0]) - np.min(vals[:, 0]) + 1)
    assert np.abs(mask.sum() - bh.area) / bh.area < 0.05 # integral mask area needs to be close to true area
    normalized_normals = bh.rnormals / np.linalg.norm(bh.rnormals, axis=1)[:, None]
    # test case 2
    for e, n in zip(bh.edges, normalized_normals):
        edge_vec = e[1] - e[0]
        assert np.all(np.abs(np.dot(edge_vec, n)) < 1.0e-8)

    # test case 3
    valsextract = np.array([[-10, 120], [90, 268], [293, 110],[40, -30]])
    bh_extract = BoundingConvexHull(valsextract)
    sinc_npx = 255
    sinc = np.sinc(np.linspace(-7, 7, sinc_npx))
    sinc2d = np.outer(sinc, sinc).reshape((1, 1, sinc_npx, sinc_npx))
    extracted_data, extracted_window_extents = BoundingConvexHull.regional_data(bh_extract, sinc2d, oob_value=np.nan)
    assert extracted_window_extents == [-10, 293, -30, 268]
    sparse_mask = np.array(bh_extract.sparse_mask)
    lines = np.hstack([bh_extract.corners, np.roll(bh_extract.corners, -1, axis=0)])
    minx = np.min(lines[:, 0:4:2]); maxx = np.max(lines[:, 0:4:2])
    miny = np.min(lines[:, 1:4:2]); maxy = np.max(lines[:, 1:4:2])
    sel = np.logical_and(np.logical_and(sparse_mask[:, 1] >= 0,
                                        sparse_mask[:, 1] < 255),
                         np.logical_and(sparse_mask[:, 0] >= 0,
                                        sparse_mask[:, 0] < 255))
    
    flat_index = (sparse_mask[sel][:, 0])*sinc_npx + (sparse_mask[sel][:, 1])
    sinc_integral = np.sum(sinc2d.ravel()[flat_index]) 
    assert np.abs(sinc_integral - np.nansum(extracted_data.ravel())) < 1.0e-8
    v = np.nanargmax(extracted_data)
    vx = v % extracted_data.shape[3]; vy = v // extracted_data.shape[3]
    cextracted = (extracted_window_extents[0] + vx,
                  extracted_window_extents[2] + vy)
    v = np.nanargmax(sinc2d)
    sincvx = v % sinc_npx; sincvy = v // sinc_npx
    csinc = tuple([sincvx, sincvy]) 
    assert csinc == cextracted
    
    # test case 4
    vals2 = np.array([[-20, -120], [0, 60], [40, -60]])
    vals3 = np.array([[-20, 58], [-40, 80], [20, 100]])
    bh2 = BoundingConvexHull(vals2)
    bh3 = BoundingConvexHull(vals3)
    assert bh.overlaps_with(bh2)
    assert not bh.overlaps_with(bh3)
    assert not bh2.overlaps_with(bh3)

    # test case 5
    assert (-1000, -1000) not in bh
    assert (30, 0) not in bh
    assert (0, 0) not in bh
    assert (-40, 30) in bh

    # test case 6
    bb = BoundingBox(-14, 20, 30, 49)
    assert bb.centre == [3, 39]
    assert bb.box_npx == (35, 20)
    assert bb.mask.shape == bb.box_npx[::-1]
    assert bb.area == 35 * 20
    
    assert np.sum(bb.mask) == bb.area
    assert (-15, 35) not in bb
    assert (0, 35) in bb

    bb2 = BoundingBoxFactory.AxisAlignedBoundingBox(bb) #enforce odd
    assert bb2.box_npx == (35, 21)
    assert bb2.area == 35 * 21
    assert (bb.sparse_mask == bb2.sparse_mask).all()
    assert (-15, 35) not in bb2
    assert (0, 35) in bb2

    bb3 = BoundingBoxFactory.AxisAlignedBoundingBox(bb, square=True) #enforce odd
    assert bb3.box_npx[0] == bb3.box_npx[1]
    assert bb3.box_npx[0] % 2 == 1 #enforce odd
    assert bb3.area == bb3.box_npx[0]**2
    assert (bb.sparse_mask == bb3.sparse_mask).all()
    assert (-15, 35) not in bb2
    assert (0, 35) in bb2

    # test case 7
    bb4s = BoundingBoxFactory.SplitBox(bb, nsubboxes=3)
    assert len(bb4s) == 9
    xlims = [(np.min(c.corners[:, 0]), np.max(c.corners[:, 0])) for c in bb4s][0:3]
    ylims = [(np.min(c.corners[:, 1]), np.max(c.corners[:, 1])) for c in bb4s][0::3]
    assert np.all(xlims == np.array([(-14, -3), (-2, 9), (10, 20)]))
    assert np.all(ylims == np.array([(30, 36), (37, 43), (44, 49)]))
    assert np.sum([b.area for b in bb4s]) == bb.area

    for bb4 in bb4s:
        assert bb4.area == np.sum(bb4.mask)

    # test case 8
    bb5 = BoundingBox(-14, 20, 30, 50)
    assert bb5.box_npx == (35, 21)
    bb6 = BoundingBoxFactory.PadBox(bb5, 41, 27)
    assert bb6.box_npx == (41, 27)
    assert bb5.centre == bb6.centre
    assert np.sum(bb5.mask) == np.sum(bb6.mask)
    bb7s = list(map(lambda x: BoundingBoxFactory.PadBox(x, 17, 11), bb4s))
    assert all([b.box_npx == (17, 11) for b in bb7s])
    assert np.sum([np.sum(b.mask) for b in bb7s]) == np.sum([np.sum(b.mask) for b in bb4s])

    # test case 9
    facet_regions = list(map(lambda f: BoundingBoxFactory.PadBox(f, 63, 63), 
                             BoundingBoxFactory.SplitBox(BoundingBoxFactory.AxisAlignedBoundingBox(bh_extract), nsubboxes=5)))
    facets = list(map(lambda pf: BoundingConvexHull.regional_data(pf, sinc2d, oob_value=np.nan),
                      facet_regions))
    stitched_image, stitched_region = BoundingBox.project_regions([f[0] for f in facets], facet_regions)
    assert np.abs(sinc_integral - np.nansum([np.nansum(f[0]) for f in facets])) < 1.0e-8
    assert np.abs(sinc_integral - np.sum(stitched_image)) < 1.0e-8
    v = np.argmax(stitched_image)
    vx = v % stitched_image.shape[3]; vy = v // stitched_image.shape[3]
    cstitched = (np.min(stitched_region.corners[:, 0]) + vx, np.min(stitched_region.corners[:, 1]) + vy)
    assert cstitched == csinc

    # test case 10
    olap_box1 = BoundingBox(110, 138, 110, 135)
    olap_box2 = BoundingBox(115, 150, 109, 150)
    olap_box3 = BoundingBox(125, 130, 125, 130)
    BoundingConvexHull.normalize_masks([olap_box1, olap_box2, olap_box3])
    ext1 = BoundingConvexHull.regional_data(olap_box1, sinc2d)[0]
    ext2 = BoundingConvexHull.regional_data(olap_box2, sinc2d)[0]
    ext3 = BoundingConvexHull.regional_data(olap_box3, sinc2d)[0]
    olaps_stitched_image, olaps_stitched_region = BoundingBox.project_regions([ext1, ext2, ext3], 
                                                                              [olap_box1, olap_box2, olap_box3])
    v = np.nanargmax(olaps_stitched_image)
    vx = v % olaps_stitched_image.shape[3]; vy = v // olaps_stitched_image.shape[3]
    cstitched_olap = (np.min(olaps_stitched_region.corners[:, 0]) + vx, 
                      np.min(olaps_stitched_region.corners[:, 1]) + vy)
    assert cstitched_olap == csinc
    assert np.abs(1.0 - np.nanmax(olaps_stitched_image)) < 1.0e-8
    
    # visual inspection
    if DEBUG:
        from matplotlib import pyplot as plt
        plt.figure(figsize=(7, 2.5))
        plt.title("Winding, normals and masking check")
        for h in [bh, bh2, bh3]:
            for ei, e in enumerate(h.edges):
                plt.plot(e[:, 0], e[:, 1], "r--")
                plt.text(e[0, 0], e[0, 1], str(ei))
        
        plt.plot(bh.edge_midpoints[:, 0], bh.edge_midpoints[:, 1], "ko")
        for e, n in zip(bh.edge_midpoints, normalized_normals):
            p0 = e 
            p = e + n*6
            plt.plot([p0[0], p[0]], [p0[1], p[1]], "b--", lw=2)
        
        plt.scatter(vals[:, 0], vals[:, 1])    
        plt.imshow(mask, extent=[np.min(vals[:, 0]), np.max(vals[:, 0]), np.max(vals[:, 1]), np.min(vals[:, 1])])

        plt.grid(True)
        plt.figure(figsize=(7, 2.5))
        plt.title("Data extraction check (global)")
        for h in [bh_extract]:
            for ei, e in enumerate(h.edges):
                plt.plot(e[:, 0], e[:, 1], "r--")
        plt.imshow(sinc2d[0, 0, :, :], extent=[0, sinc_npx, sinc_npx, 0])
        plt.grid(True)

        plt.figure(figsize=(7, 2.5))
        plt.title("Data extraction check (local)")
        for h in [bh_extract]:
            for ei, e in enumerate(h.edges):
                plt.plot(e[:, 0], e[:, 1], "r--")
        plt.imshow(extracted_data[0, 0, :, :],
            extent=[extracted_window_extents[0], extracted_window_extents[1],
                   extracted_window_extents[3], extracted_window_extents[2]])
        
        plt.figure(figsize=(7, 2.5))
        plt.title("Faceting check")
        for h in [bh_extract]:
            for ei, e in enumerate(h.edges):
                plt.plot(e[:, 0], e[:, 1], "r--")
        for f in facet_regions:
            for ei, e in enumerate(f.edges):
                plt.plot(e[:, 0], e[:, 1], "co--")
        

        plt.imshow(stitched_image[0, 0, :, :], 
            extent=[np.min(stitched_region.corners[:, 0]), np.max(stitched_region.corners[:, 0]),
                    np.max(stitched_region.corners[:, 1]), np.min(stitched_region.corners[:, 1])])

        plt.figure(figsize=(7, 2.5))
        plt.title("Overlapping faceting check")
        for f in [olap_box1, olap_box2, olap_box3]:
            for ei, e in enumerate(f.edges):
                plt.plot(e[:, 0], e[:, 1], "co--")
        

        plt.imshow(olaps_stitched_image[0, 0, :, :], 
            extent=[np.min(olaps_stitched_region.corners[:, 0]), np.max(olaps_stitched_region.corners[:, 0]),
                    np.max(olaps_stitched_region.corners[:, 1]), np.min(olaps_stitched_region.corners[:, 1])])
        plt.xlim((np.min(olaps_stitched_region.corners[:, 0]) - 15, 
                  np.max(olaps_stitched_region.corners[:, 0]) + 15))
        plt.ylim((np.min(olaps_stitched_region.corners[:, 1]) - 15, 
                  np.max(olaps_stitched_region.corners[:, 1]) + 15))
        plt.show(block=True)

if __name__ == "__main__":
    test_hull_construction()



# CubiCal: a radio interferometric calibration suite
# (c) 2017 Rhodes University & Jonathan S. Kenyon
# http://github.com/ratt-ru/CubiCal
# This code is distributed under the terms of GPLv2, see LICENSE.md for details
"""
Handles parameter databases which can contain solutions and other relevant values. 
"""
from __future__ import print_function
from builtins import range
import numpy as np
from numpy.ma import masked_array
from cubical.tools import logger

log = logger.getLogger("param_db", 0)
import scipy.interpolate, scipy.spatial
import itertools
from collections import OrderedDict


class _Record(object):
    """
    Helper class: a record is initialized from a dict, and has attributes corresponding to the 
    dict keys.
    """

    def __init__(self, **kw):
        for key, value in kw.items():
            setattr(self, key, value)


class Parameter(object):
    """
    Defines a parameter object. A parameter represents an N-dimensional set of values (and flags),
    along with axis information.

    A parameter has two types of axes: continuous and interpolatable (e.g. time, frequency), 
    and discrete (e.g. antenna, direction, correlation). Internally, a parameter is stored as 
    a set of masked arrays (e.g. 2D time/frequency arrays), one such "slice" per each discrete 
    point (e.g. per each antenna, direction, correlation, etc.)

    Discrete grids may use sets of string labels (e.g. antenna names) instead of indices.

    Each such slice may be accessed directly with get_slice() (e.g. get_slice(ant='antX',corr1=0,corr2=1))
    The returned array is a reference to the underlying slice, and may be modified. Note that each
    slice can (in principle) be defined on a different subset of the overall continuous grid.

    A parameter also supports reinterpolation onto a different continuous grid. This is slower 
    (and returns a new array). This is done via the reinterpolate() method.
    """

    def __init__(self, name, dtype, axes, interpolation_axes=[], empty=0, metadata=None, grid={}):
        """
        Initialises a Parameter object.

        Args:
            name (str):
                Name of object, e.g. "G".
            dtype (type):
                A numpy data type.
            axes (list):
                Axis names (str).
            interpolation_axes (list, optional): 
                Axes over which interpolation will be supported (1 or 2 axes).
            empty (various, optional):
                Empty value for undefined parameters, usually 0.
            metadata (str or None, optional):
                Parameter metadata.
            grid (dict, optional):
                Dict of grid coordinates, {axis: coordinates}, if known.
                Any particular axis can also be populated by add_chunk() later.
        """
        interpolation_axes = interpolation_axes or []
        assert (len(interpolation_axes) in [0, 1, 2])
        print("defining parameter '{}' over {}".format(name, ",".join(axes)), file=log(1))

        self.name, self.dtype, self.axis_labels = name, dtype, axes
        self.empty, self.metadata = empty, metadata
        # axis index: dict from axis name to axis number
        self.axis_index = {label: i for i, label in enumerate(axes)}
        # convenience member: makes axis numbers available as e.g. "self.ax.time"
        self.ax = _Record(**self.axis_index)
        # list of axis numbers which can be interpolated over (e.g. time/freq)
        self.interpolation_axes = [self.axis_index[axis] for axis in interpolation_axes]
        # list of grid values for each axis, or None if not yet defined
        self.grid = [grid.get(axis) for axis in axes]
        # index of grid values to ordinal numbers (for sparse or labelled axes)
        self.grid_index = {axis: OrderedDict([(x, i) for i, x in enumerate(values)])
                           for axis, values in grid.items() if axis not in interpolation_axes}
        # shape
        self.shape = [0 if g is None else len(g) for g in self.grid]
        # list of sets of grid values actually defined, maintained during prototype->skeleton state
        self._grid_set = [set() for _ in axes]
        # default set of grid values or labels, applicable to sparse axes
        self._default_grid_set = [set(grid.get(axis, [])) for axis in axes]

        # becomes true if parameter is populated
        # A prototype is initially unpopulated; once _update_shape has been called
        # at least once, it becomes populated
        self._populated = False

        # A Parameter object can be in one of three states, or can be transitioning between them.
        # The first two states are internal to the database; the third state is exposed to the
        # user once a database is loaded.
        #
        # * Prototype state (only __init__ has been called) describing the parameter. The grids
        #   can be completely or partially undefined at this stage.
        #
        # * Skeleton state (_update_shape/_finalize_shape has been called). The grids over the
        #   solution space are fully defined at this stage, but no values are loaded.
        #
        # * Fully populated state (_init_arrays/_paste_slice/_finalize_arrays has been called)
        #   containing a complete set of values. This is now a "user-friendly" object on which
        #   get_slice() and reinterpolate() may be called.
        #
        # Once a database is loaded, the user only sees fully populated Parameter objects. The other
        # two states only occur internally, while the database is being populated or loaded.
        #
        # The lifecycle of a Parameter in the context of a database is as follows. Keep in mind
        # that the database is simply a flat file.
        #
        # 1. A new empty database file is created. Parameters are defined (see define_param() below)
        #    one by one, and written to the file in their prototype state.
        #
        # 2. A solver runs, and spits out chunks of the solution space (see add_chunk() below).
        #    This causes _update_shape() to be called. The chunks are written to the file.
        #
        # 3. The solver finishes. _finalize_shape() is called: the Parameter is now a full
        #    skeleton. The skeletons are written to <database>.skel, and the main database file
        #    is closed. It is now what's called a "fragmented" database.
        #
        # To load a fragmented database (see PickledDatabase._load())
        #
        # 1. Skeleton Parameters are read from from <database>.skel. If this does not exist, it can
        #    be re-generated by scanning through the prototypes and slices in the main database file.
        #    _init_arrays() is called on the skeleton Parameteres.
        #
        # 2. The main database is read, and each chunk is passed to the appropriate Parameter's _paste_slice().
        #
        # 3. _finalize_arrays() is called: each Parameter is now fully populated.
        #
        # If any parameter values are changed, PickledDatabase.save() can be called to write the database in
        # "consolidated" mode. In this mode, it's simply a pickle of all fully-populated parameters.

    def _update_shape(self, shape, grid):
        """
        Called repeatedly during the prototype -> skeleton phase, as the 
        solver generates solutions for subsets of the overall parameter space.
        Updates shape of each axis based on the supplied shape and grid. Grid is a dict of 
        {axis: coordinates}, and need only be supplied for shapes that are a partial 
        slice along an axis.

        Args:
            shape ():

            grid ():

        Raises:
            ValueError:
                If grid of axis is inconsistent with previous definition.
            ValueError:
                If axis length is inconsistent with previously defined length.

        """

        self._populated = True

        for i, axis in enumerate(self.axis_labels):
            # if a grid along an axis is supplied, it is potentially a slice
            if axis in grid:
                assert (len(grid[axis]) == shape[i])
                # build up grid as we go along
                self._grid_set[i].update(grid[axis])
                # if full axis grid was predefined by define_param, and it's not
                # an interpolatable one, then supplied grid must be a subset
                if axis not in self.interpolation_axes and self.grid[i] is not None:
                    if set(grid[axis]) - self._grid_set[i]:
                        raise ValueError("grid of axis {} does not match previous definition".format(axis))
            # else it's a full axis -- check that shape conforms.
            else:
                if not self.shape[i]:
                    self.shape[i] = shape[i]
                elif self.shape[i] != shape[i]:
                    raise ValueError("axis {} of length {} does not match previously defined length {}".format(
                        axis, shape[i], self.shape[i]))
                # init grid if not already set
                self._grid_set[i] = self._default_grid_set[i]

    def _finalize_shape(self):
        """
        Finalizes shapes and axes based on accumulated _update_shape() calls.
        This turns the object into a fully fledged skeleton.

        Returns:
            bool:
                True if successful.
        """

        if not self._populated:
            return False

        self.grid_index = {}
        for iaxis, (axis, grid) in enumerate(zip(self.axis_labels, self.grid)):
            # if an actual grid was accumulated via update_shape, and either (a) axis
            # is interpolatable, or (b) no grid was predefined, then we use the accumulated grid
            if self._grid_set[iaxis] and (iaxis in self.interpolation_axes or grid is None):
                self.grid[iaxis] = grid = np.array(sorted(self._grid_set[iaxis]))
                self.shape[iaxis] = len(grid)
            # else use predefined grid, or 0...n-1 if not predefined
            elif grid is None:
                self.grid[iaxis] = grid = np.arange(self.shape[iaxis])
            # build grid index
            self.grid_index[axis] = OrderedDict([(x, i) for i, x in enumerate(grid)])
        # build interpolation grids normalized to [0,1]
        self._norm_grid = {}
        self._norm_grid_map = {}
        self._gminmax = {}
        for iaxis in self.interpolation_axes:
            grid = self.grid[iaxis]
            gmin = grid.min()
            g1 = grid - gmin
            gmax = float(g1.max()) or 1
            self._norm_grid[iaxis] = g1 = g1 / gmax
            self._gminmax[iaxis] = gmin, gmax
        print("dimensions of {} are {}".format(self.name, ','.join(map(str, self.shape))), file=log(0))
        if log.verbosity() > 0:
            for iaxis, axis in enumerate(self.axis_labels):
                if iaxis in self.interpolation_axes:
                    print("  axis {}: {} to {}".format(axis, *self._gminmax[iaxis]), file=log(1))
                else:
                    print("  axis {}: {}".format(axis, " ".join(map(str, self.grid_index[axis].keys()))), file=log(1))
        return True

    def _to_norm(self, iaxis, g):
        """ 
        Converts grid of given axis to normalized grid. 

        Args:
            iaxis ():

            g ():

        Returns:

        """

        gmin, gmax = self._gminmax[iaxis]

        return (g - gmin) / gmax

    def _from_norm(self, iaxis, g):
        """ Converts grid of given axis to unnormalized grid. 

        Args:
            iaxis ():

            g ():

        Returns:

        """

        gmin, gmax = self._gminmax[iaxis]

        return (g * gmax) + gmin

    def _init_arrays(self):
        """
        Initializes internal arrays based on skeleton information. This begins the skeleton -> 
        populated transition.
        """

        # initialize arrays -- all flagged initially, unflagged as slices are pasted in
        self._full_array = masked_array(np.full(self.shape, self.empty, self.dtype),
                                        np.ones(self.shape, bool),
                                        fill_value=self.empty)
        self._array_slices = {}
        print("  loading {}, shape {}".format(self.name, 'x'.join(map(str, self.shape))), file=log(0))

    def _paste_slice(self, item):
        """
        "Pastes" a subset of values into the internal arrays. Called repeatedly during the skeleton 
        -> populated transition.

        Args:
            item ():

        """

        # form up slice operator to "paste" slice into array
        array_slice = []
        for iaxis, axis in enumerate(self.axis_labels):
            if axis in item.grid:
                grid_index = self.grid_index[axis]
                array_slice.append(np.array([grid_index[g] for g in item.grid[axis]]))
            else:
                array_slice.append(np.arange(self.shape[iaxis]))
        self._full_array[np.ix_(*array_slice)] = item.array

    def _finalize_arrays(self):
        """
        Finalizes internal arrays by breaking them into slices. This completes the skeleton -> 
        populated transition.
        """

        interpol_axes = self.interpolation_axes  # list of axis numbers over which we interpolate
        interpol_shape = []  # shape of interpolatable slice
        slicer_axes = []  # list of axis numbers over which we slice
        slicers = []  # list of iterators for each sliced axis
        for i, shape in enumerate(self.shape):
            if i in interpol_axes:
                interpol_shape.append(shape)
                slicers.append((None,))
            else:
                slicer_axes.append(i)
                slicers.append(range(shape))

        self._interpolators = {}

        # get grid over interpolatable axes
        print("decomposing {} into slices".format(self.name), file=log(2))
        # loop over all not-interpolatable slices (e.g. direction, antenna, correlation)
        for slicer in itertools.product(*slicers):
            array_slicer = tuple([slice(None) if sl is None else sl for sl in slicer])
            array = self._full_array[array_slicer]
            flags = array.mask
            # this is a list, per axis, of which subset of the full axis grid the slice is associated with.
            # slice(None) is the full grid (i.e. nothing masked in the array)
            grids = ([self.grid[axis] for axis in interpol_axes],
                     [self._norm_grid[axis] for axis in interpol_axes])
            subset = [slice(None) for _ in interpol_axes]
            if flags is not np.ma.nomask:
                # now, for every axis in the slice, cut out fully flagged points
                allaxis = set(range(array.ndim))
                for iaxis in range(array.ndim):
                    # find points on this axis which are fully flagged along other axes
                    if array.ndim == 1:
                        allflag = flags
                    else:
                        allflag = flags.all(axis=tuple(allaxis - {iaxis}))
                    # all flagged? Indicate this by array=None
                    if allflag.all():
                        print("  slice {} fully flagged".format(slicer), file=log(2))
                        array = None
                        break
                    # if such points exist, extract subset of array and grid
                    elif allflag.any():
                        print("  slice {} flagged at {} {} points".format(slicer, allflag.sum(),
                                                                                    self.axis_labels[
                                                                                        interpol_axes[iaxis]]), file=log(2))
                        # make corresponding slice
                        array_slice = [slice(None)] * array.ndim
                        # also set subset to the mask of the valid points
                        array_slice[iaxis] = subset[iaxis] = ~allflag
                        for gr in grids:
                            gr[iaxis] = gr[iaxis][subset[iaxis]]
                        # apply it to array, flags
                        array = array[tuple(array_slice)]
            # store resulting slice. E.g. for 2 interpolation axes of shape N and M, the
            # attributes are:
            #   array:      shape NxM
            #   grid:       [N-vector, M-vector] grid points
            #   norm_grid:  [N-vector, M-vector] normalized grid points
            #   subset:     [N-vector bool or slice(None), M-vector bool or slice(None)]
            #   gridmap:    [{x: i}, {y: j}] two dicts giving reverse mapping from grid values (of interpolatable axes)
            #               to rows/columns of array
            self._array_slices[tuple(slicer)] = _Record(array=array,
                        grid=grids[0], norm_grid=grids[1], subset=subset,
                        gridmap=[{x: i for i, x in enumerate(grid)} for grid in grids[0]])
        self._full_array = None

    def _get_slicer(self, **axes):
        """  Builds up an index tuple corresponding to keyword arguments that specify a slice. """

        slicer = []
        for iaxis, (axis, n) in enumerate(zip(self.axis_labels, self.shape)):
            if axis in axes:
                value = value0 = axes[axis]
                # if value of axis is in the grid index, look it up
                if axis in self.grid_index and value in self.grid_index[axis]:
                    value = self.grid_index[axis][value]
                # better be an int now
                if not isinstance(value, int):
                    raise ValueError("invalid axis {}={}".format(axis, value0))
                slicer.append(value)
            elif iaxis in self.interpolation_axes:
                slicer.append(None)
            elif n == 1:
                slicer.append(0)
            else:
                raise TypeError("axis {} not specified".format(axis))

        return tuple(slicer)

    def get_slice(self, **axes):
        """
        Returns array and grids associated with given slice, as given by keyword arguments.
        Note that a single index must be specified for all discrete axes with a size of greater 
        than 1. Array may be None, to indicate no solutions for the given slice.
        """

        rec = self._array_slices[self._get_slicer(**axes)]

        return rec.array, rec.grid

    def is_slice_valid(self, **axes):
        """
        Returns True if there are valid solutions for a given slice, as given by keyword arguments.
        Note that a single index must be specified for all discrete axes with a size of greater 
        than 1.
        """

        rec = self._array_slices[self._get_slicer(**axes)]

        return rec.array is not None

    def get_cube(self):
        """
        Returns full cube of solutions (dimensions of all axes), interpolated onto the superset of 
        all slice grids.
        """
        return self.lookup(grid={self.axis_labels[iaxis]: self.grid[iaxis]
                                 for iaxis in self.interpolation_axes})


    def _prepare_interpolation(self, **grid):
        """
        Helper function, to interpolate a named parameter onto the specified grid.

        Args:
            grid (dict): 
                Axes to be returned. For interpolatable axes, grid should be a vector of coordinates
                (the superset of all slice coordinates will be used if an axis is not supplied). 
                Discrete axes may be specified as a single value, or a list of values, or a
                slice object. If any axis is missing, the full axis is returned.

        Returns:
            A tuple of lists:
                - output_shape: shape of output array (before output_reduction is applied)
                - input_slicers: per each axis, an iterable giving the points to be sampled along that axis, or
                       [None] for the interpolatable axes. The net result is that
                       itertools.product(*input_slicers) iterates over all relevant slices
                - output_slicers: corresponding list of iterables giving the index in the output array to which
                       the resampled input slice is to be assigned
                - output_slice_grid: per each interpolatable axis with size>1, grid onto which interpolation is done,
                                     or None if size==1
                - output_reduction:  index applied to output array, to eliminate axes for which only a single element
                        was requested
        """

        output_shape = []
        input_slicers = []
        output_slicers = []
        output_slice_grid = []
        output_reduction = []

        for iaxis, (axis, size) in enumerate(zip(self.axis_labels, self.shape)):
            output_reduction.append(0 if axis in grid and np.isscalar(grid[axis]) else slice(None))
            # interpolatable axis
            if iaxis in self.interpolation_axes:
                if axis in grid:
                    g = np.array(grid[axis]) if np.isscalar(grid[axis]) else grid[axis]
                else:
                    g = self.grid[iaxis]
                output_slice_grid.append(g)
                output_shape.append(len(g))
                input_slicers.append([None])
                output_slicers.append([slice(None)])
            # discrete axis, so return shape is determined by index in **grid, else full axis returned
            else:
                sl = np.arange(size)
                if axis in grid:
                    value = grid[axis]
                    # list of values: looked up in grid index
                    if isinstance(value,(list, tuple, np.ndarray)):
                        sl = [self.grid_index[axis][x] for x in value]
                    # slice: direct index
                    elif type(value) is slice:
                        sl = sl[grid[axis]]
                    # other: look up in grid index
                    else:
                        sl = self.grid_index[axis].get(value, None)
                        if sl is None:
                            raise ValueError("invalid axis value {}={}".format(axis, value))

                if np.isscalar(sl):
                    output_shape.append(1)
                    input_slicers.append([sl])
                    output_slicers.append([0])
                else:
                    sl = np.array(sl)
                    output_shape.append(len(sl))
                    input_slicers.append(sl)
                    output_slicers.append(sl - sl[0])

        return output_shape,\
                input_slicers,\
                output_slicers,\
                output_slice_grid,\
                output_reduction


    def reinterpolate(self, **grid):
        """
        Interpolates named parameter onto the specified grid.

        Args:
            grid (dict): 
                Axes to be returned. For interpolatable axes, grid should be a vector of coordinates
                (the superset of all slice coordinates will be used if an axis is not supplied). 
                Discrete axes may be specified as a single index, or a vector of indices, or a 
                slice object. If any axis is missing, the full axis is returned.

        Returns:
            :obj:`~numpy.ma.core.MaskedArray`:
                Masked array of interpolated values. Mask will indicate values that could not be 
                interpolated. Shape of masked array will correspond to the order axes defined by 
                the parameter, omitting the axes in \*\*grid for which only a single index has been
                specified.
        """

        output_shape,\
        input_slicers,\
        output_slicers,\
        output_slice_grid,\
        output_reduction    = self._prepare_interpolation(**grid)

        # create output array of corresponding shape
        output_array = np.full(output_shape, self.empty, self.dtype)

        print("will interpolate {} solutions onto {} grid".format(self.name,
                                "x".join(map(str, output_shape))), file=log(1))

        # now loop over all slices
        for slicer, out_slicer in zip(itertools.product(*input_slicers), itertools.product(*output_slicers)):
            # arse is the current array slice we work with
            arse = self._array_slices[slicer]
            if arse.array is None:
                print("  slice {} fully flagged".format(slicer), file=log(2))
                output_array[out_slicer] = np.nan
            else:
                # Check which subset of the slice needs to be interpolated
                # We build up the following lists describing the interpolation process
                # (assuming N interpolatable axes, of which M<=N are of size>1, so interpolation
                # is done in M dimensions. For now, M=0,1,2)
                #
                # segment_grid (M): float array of normalized coordinates corresponding
                #                   to (input) array segment being interpolated
                # output_coord (M): float array of normalized (output) interpolation coordinates
                # array_segment_slice (N): index object to extract segment from array. Note that this
                #     will reduce dimensions to M
                # interp_shape (M): shape of interpolation result (M-dim)
                # interp_broadcast (N): index object to broadcast result of interpolation (M-dim)
                #     back to N-dim shape
                # input_grid_segment (M): used to identify the interpolator object in the cache
                segment_grid = []
                output_coord = []
                array_segment_slice = []
                input_grid_segment = []
                interp_shape = []
                interp_broadcast = []
                # build up the two objects above
                for iaxis, outgr, agr, angr in zip(self.interpolation_axes,
                                                   output_slice_grid,
                                                   arse.grid, arse.norm_grid):
                    # interpolatable axis of size >1: process by interpolator
                    if self.grid[iaxis].size > 1:
                        output_coord.append(self._to_norm(iaxis, outgr))
                        # find [i0,i1]: index of envelope of output grid points in the array grid
                        i0, i1 = np.searchsorted(agr, [outgr[0], outgr[-1]])
                        i0, i1 = max(0, i0 - 1), min(len(agr), i1 + 1)
                        segment_grid.append(angr[i0:i1])
                        input_grid_segment.append((iaxis, i0, i1))
                        # extract segment on input to interpolator, do not broadcast back out
                        array_segment_slice.append(slice(i0, i1))
                        interp_shape.append(len(outgr))
                        interp_broadcast.append(slice(None))
                    # size=1: reduce it in the input to interpolator, broadcast back out in the output with newaxis
                    else:
                        array_segment_slice.append(0)
                        interp_broadcast.append(np.newaxis)

                # see if we can reuse an interpolator
                interpolator, input_grid_segment0 = self._interpolators.get(slicer, (None, None))

                # check if the grid segment  of the cached interpolator is a strict superset of the current one: if so,
                # we can reuse the interpolator. Otherwise, create a new one
                if not interpolator or len(input_grid_segment0) != len(input_grid_segment) or \
                        not all([ia == ja and i0 <= j0 and i1 >= j1
                                 for (ia, i0, i1), (ja, j0, j1) in zip(input_grid_segment0, input_grid_segment)]):
                    print("  slice {} preparing {}D interpolator for {}".format(slicer,
                        len(segment_grid), ",".join(["{}:{}".format(*seg[1:]) for seg in input_grid_segment])), file=log(2))
                    # arseg: relevant segment of array slice
                    arseg = arse.array[tuple(array_segment_slice)]
                    # arav: linear array of all values, adata: all unflagged values
                    arav = arseg.ravel()
                    adata = arav.data[~arav.mask] if arav.mask is not np.ma.nomask else arav.data

                    # this is used by the 2D and >2D cases, so make a function
                    def makeLinearNDInterpolator(segment_grid, arav, adata):
                        meshgrids_full = np.array([g.ravel() for g in np.meshgrid(*segment_grid, indexing='ij')]).T
                        meshgrids = meshgrids_full[~arav.mask, :] if arav.mask is not np.ma.nomask else meshgrids_full
                        qhull_options = "Qbb Qc Qz Q12"
                        # edge case of <4 valid points. Delaunay falls over, so artificially duplicate points,
                        # and allow Qhull juggling with the QJ option
                        if len(adata) < 4:
                            adata = np.resize(adata, 4)
                            meshgrids = np.resize(meshgrids, (4, 2))
                            qhull_options += " QJ"
                        # edge case of all points along an axis being on the same line. Allow juggling then,
                        # else Delaunay also falls over
                        elif len(set(meshgrids[:, 0])) < 2 or len(set(meshgrids[:, 1])) < 2:
                            qhull_options += " QJ"
                        triang = scipy.spatial.Delaunay(meshgrids, qhull_options=qhull_options)
                        return scipy.interpolate.LinearNDInterpolator(triang, adata, fill_value=np.nan), meshgrids_full

                    # edge case: no valid data. Make fake interpolator
                    if not len(adata):
                        interpolator = lambda coords: np.full(coords.shape[:-1], np.nan, adata.dtype)
                    # for ndim=0, just return the 0,0 element of array
                    elif not len(segment_grid):
                        interpolator = lambda coords: arse.array
                    # for ndim=1, use 1D interpolator
                    elif len(segment_grid) == 1:
                        agrid = segment_grid[0]
                        if arav.mask is not np.ma.nomask:
                            agrid = agrid[~arav.mask]
                        # handle edge case of 1 valid point, since interp1d() falls over on this
                        if len(adata) == 1:
                            adata = np.array([adata[0], adata[0]])
                            agrid = np.array([agrid[0] - 1e-6, agrid[0] + 1e-6])
                        # make normal interpolator
                        interpolator = scipy.interpolate.interp1d(agrid, adata, bounds_error=False,
                                                                  fill_value="extrapolate")
                    # for 2D, use fill, then interp2d (which extrapolates)
                    elif len(segment_grid) == 2:
                        # if any points inside the slice are masked, inpaint them with LinearNDInterpolator first,
                        # so that we have a complete 2D filled_array
                        if arseg.mask is not np.ma.nomask and arseg.mask.any():
                            ndinp, meshgrids = makeLinearNDInterpolator(segment_grid, arav, adata)
                            # list of coordinates where we're missing values
                            missing_grids = meshgrids[arav.mask, :]
                            # fill missing values using interpolator
                            filled_array = arav.copy()
                            filled_array[arav.mask] = ndinp(missing_grids)
                            filled_array = filled_array.reshape(arseg.shape)
                        else:
                            filled_array = arseg.data
                        # make 2D interpolator
                        interpolator1 = scipy.interpolate.interp2d(segment_grid[0], segment_grid[1],
                                                                  filled_array.T, bounds_error=False)
                        interpolator = lambda *output_coord: interpolator1(*output_coord).T

                    # for >2 dims, use LinearNDInterpolator. Note that this does not extrapolate.
                    elif len(segment_grid) > 2:
                        interpolator1, _ = makeLinearNDInterpolator(segment_grid, arav, adata)
                        interpolator = lambda *output_coord: \
                            interpolator1(np.array([x.ravel() for x in np.meshgrid(*output_coord, indexing='ij')]).T).\
                                reshape(interp_shape)

                    # cache the interpolator
                    self._interpolators[slicer] = interpolator, input_grid_segment

                result = interpolator(*output_coord)

                # # make a meshgrid of output coordinates, and massage into correct shape for interpolator
                # coords = np.array([x.ravel() for x in np.meshgrid(*output_coord, indexing='ij')])
                # # call interpolator. Reshape into output slice shape
                # result = interpolator(coords.T).reshape(interp_shape)
                print("  interpolated onto {} grid".format("x".join(map(str, interp_shape))), file=log(2))
                output_array[out_slicer] = result[tuple(interp_broadcast)]
        # return array, throwing out unneeded axes
        output_array = output_array[tuple(output_reduction)]
        # also, mask missing values from the interpolator with the fill value
        missing = np.isnan(output_array)
        output_array[missing] = self.empty
        print("{} solutions: interpolation results in {}/{} missing values".format(self.name,
                            missing.sum(), missing.size), file=log(1))
        return masked_array(output_array, missing, fill_value=self.empty)


    def lookup(self, **grid):
        """
        Looks up values on the specified grid. The difference with reinterpolate() is that
        the grid values must match exactly. Any missing values will be masked out.

        Args:
            grid (dict): 
                Axes to be returned. For interpolatable axes, grid should be a vector of coordinates
                (the superset of all slice coordinates will be used if an axis is not supplied). 
                Discrete axes may be specified as a single index, or a vector of indices, or a 
                slice object. If any axis is missing, the full axis is returned.

        Returns:
            :obj:`~numpy.ma.core.MaskedArray`:
                Masked array of values. Mask will indicate grid points that didn't match.
        """
        output_shape,\
        input_slicers,\
        output_slicers,\
        output_slice_grid,\
        output_reduction    = self._prepare_interpolation(**grid)

        # create output array of corresponding shape
        output_array = np.full(output_shape, self.empty, self.dtype)
        output_mask  = np.ones(output_shape, bool)

        print("will lookup {} solutions on {} grid".format(self.name,
                                "x".join(map(str, output_shape))), file=log(1))


        # now loop over all slices
        for slicer, out_slicer in zip(itertools.product(*input_slicers), itertools.product(*output_slicers)):
            # arse is the current array slice we work with
            arse = self._array_slices[slicer]
            if arse.array is None:
                print("  slice {} fully flagged".format(slicer), file=log(2))
            else:
                # segment_grid: float array of normalized coordinates corresponding
                #               to segment being interpolated over
                # output_coord: float array of normalized output coordinates
                # array_segment_slice: index object to extract segment from array. Note that this
                #     may also reduce dimensions, if size=1 interpolatable axes are present.
                # input_grid_segment: used to cache the interpolator.
                input_indices = []
                output_indices = []
                # build up the two objects above
                for outgr, gmap in zip(output_slice_grid, arse.gridmap):
                    # for each output grid point, lookup corresponding grid point in the array slice
                    ij = [ (i, gmap.get(x)) for i,x in enumerate(outgr) ]
                    input_indices.append([ j for i,j in ij if j is not None])
                    output_indices.append([ i for i,j in ij if j is not None])
                print("  slice {}: looking up {} valid points".format(slicer,
                        "x".join([str(len(idx)) for idx in input_indices])), file=log(2))

                out = output_array[out_slicer]
                outmask = output_mask[out_slicer]
                ox, ix = np.ix_(*output_indices), np.ix_(*input_indices)
                out[ox] = arse.array[ix]
                if arse.array.mask is np.ma.nomask:
                    out[ox] = False
                else:
                    outmask[ox] = arse.array.mask[ix]

        # return array, throwing out unneeded axes
        output_array = output_array[tuple(output_reduction)]
        output_mask  = output_mask[tuple(output_reduction)]
        output_array[output_mask] = self.empty

        print("{} solutions: no interpolation done on {}/{} missing values".format(self.name,
                            output_mask.sum(), output_mask.size), file=log(1))

        return masked_array(output_array, output_mask, fill_value=self.empty)

    def find_mismatched_grids(self, **grid):
        """
        Looks up the specified grid values and returns list of mismatches.
        Empty list indicates that lookup() can be done on the grids
        (if not empty, then reinterpolate() should be used).

        Args:
            grid (dict): 
                Grid coordinates to look up

        Returns:
            list of (axis, list_of_mismatching_values) tuples
        """
        mismatches = []
        for axis, gridvalues in grid.items():
            iaxis = self.axis_index[axis]
            grid0values = self._grid_set[iaxis] or set(range(self.shape[iaxis]))
            mismatch = set(gridvalues) - grid0values
            if mismatch:
                mismatches.append((axis, sorted(mismatch), iaxis in self.interpolation_axes))

        return mismatches

    def release_cache(self):
        """ 
        Scrubs the object in preparation for saving to a file (e.g. removes cache data etc.)
        """
        self._interpolators = {}
